/* Stuff generic to all Ethernet controllers
 * Copyright 1991 Phil Karn, KA9Q
 */
#include "global.h"
#ifdef 80211
#include "mbuf.h"
#include "iface.h"
#include "arp.h"
#include "80211.h"
#include "enet.h"

#if !defined(_lint)
static char rcsid[] OPTIONAL = "$Id: 80211.c,v 1.1.1.1 2004/02/09 21:47:54 blantz Exp $";
#endif

static struct mbuf *hton80211(struct ether *ether,struct mbuf *data);



/* Convert 80211 header in host form to network mbuf */
static struct mbuf *
hton80211(
struct hdr_80211_t *hdr,
struct mbuf *data
){
char *cp;
struct mbuf *bp;

    if(data == NULL || (bp = pushdown (data, 80211_HDR_LEN)) == NULLBUF)
	return NULLBUF;

    cp = (char *) bp->data;

    (void) put16 ((unsigned char *) cp, (int16) hdr->fc);
    (void) put16 ((unsigned char *) cp, (int16) hdr->dur);
    memcpy (cp, hdr->a1, 80211_ADDR_LEN);
    cp += 80211_ADDR_LEN;
    memcpy (cp, hdr->a2, 80211_ADDR_LEN);
    cp += 80211_ADDR_LEN;
    memcpy (cp, hdr->a3, 80211_ADDR_LEN);
    cp += 80211_ADDR_LEN;
    (void) put16 ((unsigned char *) cp, (int16) hdr->seq);

    return bp;
}



/* Extract 80211 header */
int
ntoh80211(
struct hdr_80211_t *hdr,
struct mbuf **bpp
){
    
    hdr->fc = pull16 (bpp);
    hdr->dur = pull16 (bpp);
    (void) pullup (bpp, hdr->a1, EADDR_LEN);
    (void) pullup (bpp, hdr->a2, EADDR_LEN);
    (void) pullup (bpp, hdr->a3, EADDR_LEN);
    hdr->seq = pull16 (bpp);
    return 80211_HDR_LEN;
}



/* Send an IP datagram on 80211 */
int
enet_80211_send(
struct mbuf *bp,	/* Buffer to send */
struct iface *iface,	/* Pointer to interface control block */
uint32 gateway,		/* IP address of next hop */
int prec OPTIONAL,
int del OPTIONAL,
int tput OPTIONAL,
int rel OPTIONAL
){
char *egate;

    if (gateway == iface->broadcast) /* This is a broadcast IP datagram */
        return (*iface->output) (iface, (char *) Ether_bdcst, iface->hwaddr, 80211_FTYPE_DATA, bp);

    /* ??? */
    egate = res_arp (iface, ARP_ETHER, gateway, bp);
    if(egate != NULLCHAR)
	return (*iface->output) (iface, egate, iface->hwaddr, 80211_FTYPE_DATA, bp);
    return 0;
}



/* Send a packet with 80211 header */
int
enet_80211_output(
struct iface *iface,	/* Pointer to interface control block */
const char *dest,	/* Destination Ethernet address */
char *source,		/* Source Ethernet address */
int16 type,		/* Type field */
struct mbuf *data	/* Data field */
){
struct buf_80211_t pk;
struct mbuf *bp;

    memcpy (pk.a1, dest, EADDR_LEN);
    memcpy (pk.a2, source, EADDR_LEN);
    pk.fc = type;
    pk.dur = 0;
    pk.seq = 0;
    if ((bp = hton80211(&pk,data)) == NULLBUF)	{
	free_p (data);
	return -1;
    }
    return (*iface->raw)(iface,bp);
}



/* Process incoming 80211 packets. Shared by all wireless drivers. */
void
eproc_80211(
struct iface *iface,
struct mbuf *bp
){
struct hdr_80211_t hdr;

	/* Remove 80211 header and kick packet upstairs */
	(void) ntoh80211(&hdr,&bp);

    /* ??? */
#if 0
    if(memcmp (hdr.dest, iface->hwaddr, EADDR_LEN) && memcmp (hdr.dest, Ether_bdcst, EADDR_LEN))	{
	free_p(bp);
	return;
    }
#endif

    switch(hdr.type)	{
	case 80211_FTYPE_DATA:
		(void) ip_route (iface, bp, /*hdr.dest[0] & 1*/ 0);
		break;
	default:
		free_p (bp);
		break;
    }
}

#endif		/* 80211 */
