
/*
 * APRS Services for JNOS 111x, TNOS 2.30 & TNOS 2.4x
 *
 * April,2001-February,2002 - Release # 1.11.2
 *
 * Designed and written by VE4KLM, Maiko Langelaar
 *
 * For non-commercial use (Ham Radio) only !!!
 *
 */

#include <ctype.h>

#include "global.h"

#ifdef	APRSD

#include "mbuf.h"

/* 01Feb2002, Maiko, DOS compile for JNOS will complain without these */
#ifdef	MSDOS
#include "iface.h"
#include "arp.h"
#include "slip.h"
#endif

#include "ax25.h"
#include "udp.h"

#include "aprs.h"	/* 16May2001, VE4KLM, Added prototypes */

#if !defined(_lint)
static char rcsid[] OPTIONAL = "$Id: aprs.c,v 1.1.1.1 2004/02/09 21:47:55 blantz Exp $";
#endif

static char data_APZxxx[300];	/* make sure we have 'lots' of room */

extern unsigned long internal_hostid;	/* the only kludge so far */

/* 19Jul2001, VE4KLM, We should drop our own packets coming from a digi */
extern char logon_callsign[15];

extern int aprs_poe;	/* 05Sep2001, Maiko, Flag Point of Entry call */

/*
 * 30Aug2001, Maiko, New function to format the POE (point of entry) callsign,
 * so that packets sent to the internet APRS system will now be traceable to
 * the gateway that received the packets and injected them into the system.
 *
 * 05Sep2001, Maiko, Even though not everyone is in agreement, those that I
 * see as the major players seem to agree that '-0n' is at least backwards
 * compatible and will do no harm to the existing system. I will now make
 * this an option (ie, sysops can switch off if needed).
 */

static char *format_POE_call (char *ptr)
{
	char *lptr = logon_callsign;

	int ssid = 0;

	*ptr++ = ',';

	while (*lptr && *lptr != '-')
		*ptr++ = *lptr++;

	if (*lptr++ == '-')
		ssid = atoi (lptr) % 10;

	/* 30Aug2001, Maiko, Stick with generic (-Gn) gateway for now */
	/* 05Sep2001, Maiko, Switched to using (-0n), leading 0 in SSID */
	ptr += sprintf (ptr, "-0%d", ssid);
	
	return ptr;
}

int aprs_send (char *data, int len)
{
	struct mbuf *bp;
	struct socket lsock,fsock;

	lsock.address = INADDR_ANY;
	lsock.port = 1315;

	/*
	 * 30May2001, VE4KLM, If this is not set, then set it to the address
	 * of this TNOS box. This basically means that the 'aprs internal'
	 * entry in autoexec.nos is no longer required except in special cases.
	 * ALOT of TNOS systems now have linux doing the encap and tunnel, so
	 * for those systems, there is no longer any need for 'aprs internal'.
	 */
	if (internal_hostid == 0)
		internal_hostid = Ip_addr;

	fsock.address = internal_hostid;	/* kludge for now */
	fsock.port = 93;

	if ((bp = alloc_mbuf (len)) == NULLBUF)
		return -1;

	strncpy ((char*)bp->data, data, len);

	bp->cnt = (int16)len;

	send_udp (&lsock, &fsock, 0, 0, bp, 0, 0, 0);

	return 0;
}

static char aprs_dest[2] = { 'A' << 1, 'P' << 1 };
static char beac_dest[2] = { 'B' << 1, 'E' << 1 };

/* 17May2001, Maiko, Oops forgot to add ID as a possible dest call */
static char id_dest[2] = { 'I' << 1, 'D' << 1 };

/*
 * 20Aug2001, VE4KLM, Put in NOGATE and RFONLY digi calls, decided to move
 * the optout of igate code into this module instead. That way I can keep
 * stats on who opted out, as well, there's no point sending to the main
 * server just to be discarded anyways
 */
static char optout1[AXALEN] = { 'N' << 1, 'O' << 1, 'G' << 1, 'A' << 1, 'T' << 1, 'E' << 1, '0' << 1 };
static char optout2[AXALEN] = { 'R' << 1, 'F' << 1, 'O' << 1, 'N' << 1, 'L' << 1, 'Y' << 1, '0' << 1 };

/*
 * 20Aug2001, VE4KLM, New function to quickly determine whether passed
 * digicall (internal format) is one of the optout callsigns
 */
static int optout (char *digicall)
{
	if (addreq (digicall, optout1))
		return 1;

	if (addreq (digicall, optout2))
		return 1;

	return 0;
}

/*
 * 20Aug2001, VE4KLM, Moved the digi processing code into a function instead,
 * because I had duplicate copies of this code in aprs_processing function,
 * so now we just call this function instead.
 */

static char *process_digis (struct ax25 *hdr, char *ptr)
{
	register int cnt;

	char tmp[AXBUF];

	if (hdr->ndigis)
	{
		for (cnt = 0; cnt < hdr->ndigis; cnt++)
		{
			/* 20Aug2001, VE4KLM, Check if user wants to opt out */
			if (optout (hdr->digis[cnt]))
			{
				if (aprs_debug)
				{
					aprslog (-1, "User [%s] opted out",
						pax25 (tmp, hdr->source));
				}

				/* 20Aug2001, VE4KLM, Flag last stat as an optout */
				set_last_optout ();

				return (char*)0;	/* this indicates an OPT OUT !!! */
			}

			/* 05Sep2001, VE4KLM, Better do this right, and actually check
			 * to see if the repeated BIT is set in the digi list
			 *
			ptr += sprintf (ptr, ",%s", pax25 (tmp, hdr->digis[cnt]));
			 *
			 */

			ptr += sprintf (ptr, ",%s%s", pax25 (tmp, hdr->digis[cnt]),
 					(hdr->digis[cnt][ALEN] & REPEATED) ? "*" : "");
		}

		/*
		 * the last digi needs an asterisk to identify it
		 * 05Sep2001, Maiko, This is not required anymore, since the above
		 * check for the H (repeated) BIT should take care of it now.
		 *
		ptr += sprintf (ptr, "*");
	 	 *
		 */
	}

	return ptr;
}

/*
 * Main APRS Frame Processing Call - called from within the ax25.c module
 */

int aprs_processing (struct ax25 *hdr, struct iface *iface, struct mbuf *bp)
{
	char tmp[AXBUF], *ptr, *ptr2;

	int len;

#ifdef	APRS_MIC_E

	char miceout[200];

	/*
	 * 15Jun2001, VE4KLM, Developing MIC-E support, new
	 * function, aprs_mice, to decode the destination call
	 * and user data, to see if it is indeed MIC-E info.
	 */
	if (aprs_mice (hdr->dest, bp->data, miceout))
	{
		/* don't forget to update the heard table */
		/* 18Jun2001, VE4KLM, don't forget to skip control and pid characters */
		/* 16Aug2001, Maiko, Changed parameter list, more info to use now */
		update_aprshrd (hdr, iface, *(bp->data + 2));

		ptr = data_APZxxx;

		/* Replace the MIC-E dest call with this IGATE's dest call */
		/* 20Jun2001, VE4KLM, Unproto call now defined in aprs.h */
		ptr += sprintf (ptr, "%s>%s", pax25 (tmp, hdr->source), APRSDC);

		/*
	 	 * 20Aug2001, VE4KLM, Unproto path processing now in it's own
		 * function (to cut code duplication). IF a NULL is returned,
		 * then it means user opted out or something else.
		 */
		if ((ptr = process_digis (hdr, ptr)) == (char*)0)
			return 1;

		/* 30Aug2001, Maiko, Now have a function to insert the POE call,
		 * however this is now on hold. See comments further down in this
		 * code regarding the POE stuff.
		 *
		 * 05Sep2001, Maiko, Decided to add this, since there has been
		 * some agreement on backwards compatibility. It's not 100 percent
		 * agreed, but the major players don't seem to mind using '-0N', so
		 * I'll make it an option for now. What is really unfortunate is the
		 * fact that the APRS Internet System has been around for some time
		 * now, and *still* they don't have an official APRS IS Spec :-(
		 *
		 * If sysop does not want POE injection, it can be switched off.
		 */
		if (aprs_poe)
			ptr = format_POE_call (ptr);

		/*
		 * 18Jun2001, VE4KLM, Better add the newline and make sure
		 * the string terminator is also sent to the world wide APRS
		 * network.
		 */
		ptr += sprintf (ptr, ":%s\n", miceout);

		/* use UDP to send it to myself (I use UDP to talk inside TNOS) */
		aprs_send (data_APZxxx, strlen (data_APZxxx) + 1);

		return 1;
	}

#endif

	/*
	 * At present - only allow packets destined for callsigns 'BEACON'
	 * and 'ID', or any callsign beginning with the letters 'AP'.
	 */
	if ((hdr->dest[0] == aprs_dest[0] && hdr->dest[1] == aprs_dest[1]) ||
		(hdr->dest[0] == beac_dest[0] && hdr->dest[1] == beac_dest[1]) ||
		(hdr->dest[0] == id_dest[0] && hdr->dest[1] == id_dest[1]))
	{
		/* If we get in here, then we have a possible APRS frame */

		ptr2 = bp->data + 2;	/* skip ax25 ctl and pid fields */

		len = bp->cnt - 2;

		/*
		 * Now validate the Data Type Identifier to see if this is
		 * really an APRS frame. Note, not all possible identifiers
		 * may be supported by this TNOS APRS Server implementation.
		 * 17May2001, VE4KLM, Now pass pointer to 'dti_is_valid ()'.
		 * Before I simply passed the DTI char, but now I need to
		 * pass the entire data string for stronger data checking.
		 * 09Jan2002, New DTI validation now in aprssrv module.
		 */

		if (valid_dti (*ptr2))
		{
			/* 11Jan2002, Maiko, Attempt at better DTI data validation */
			if (!valid_dti_data (*ptr2, ptr2 + 1, len - 1))
				return 1;

			/*
			 * 19Jul2001, VE4KLM, Oops - If we get a packet with our source
			 * call, then likely it happened because some local digipeater
			 * repeated it and we picked it up. Definitely ignore it !! And
			 * you never know, it could be an imposter (however unlikely).
			 */
			if (!strcmp (pax25 (tmp, hdr->source), logon_callsign))
			{
				if (aprs_debug)
					aprslog (-1, "dropping a packet originated by us");

				return 1;
			}

            /* 14May2001, VE4KLM, Now track User stats */
			/* 16Aug2001, Maiko, Changed parameter list, more info to use now */
            update_aprshrd (hdr, iface, *ptr2);

			/*
			 * If we get in here, then it's almost for sure an APRS
			 * frame. Now build a full message to send to an aprs.net
			 * server - prepend the data with a source path header,
			 * and send out via internal UDP to our TNOS server.
		 	 */

			ptr = data_APZxxx;

			ptr += sprintf (ptr, "%s>", pax25 (tmp, hdr->source));

			ptr += sprintf (ptr, "%s", pax25 (tmp, hdr->dest));

			/*
		 	 * 20Aug2001, VE4KLM, Unproto path processing now in it's own
			 * function (to cut code duplication). IF a NULL is returned,
			 * then it means user opted out or something else.
			 */
			if ((ptr = process_digis (hdr, ptr)) == (char*)0)
				return 1;

			/*
			 * 29Aug2001, VE4KLM, Experimental insertion of MYCALL-Gn so that
			 * the packet can now be traced to which IGATE was used to inject
			 * it into the internet APRS system. Discussed at length with Bob
		 	 * Bruninga (WB4APR) and others in the APRS community, BUT ...
			 *
			 * 30Aug2001, I have now put this idea on hold because it seems
			 * the debate is still going on, and I don't see a light to the
			 * end of the tunnel yet... The latest proposal is now to have a
			 * leading 0 (zero) in the SSID which would keep the POE callsign
			 * valid. Waiting for the outcome (if there is one). In the mean
			 * time, I'll have to keep the POE out of the loop for now :-(
			 *
			 * 30Aug2001, Maiko, Created a function to insert the POE call
			 *
		 	 * 05Sep2001, Maiko, Decided to start using this as an option,
			 * please see further comments I have scribbled in at the other
			 * call made to the format_POE_call () function earlier on in
			 * this module.
			 *
			 * If sysop does not like POE injection, it can be switched off.
			 */
			if (aprs_poe)
				ptr = format_POE_call (ptr);

			ptr += sprintf (ptr, ":%.*s\n", len, ptr2);

			len = ptr - data_APZxxx + 1;	/* include the NULL for internet */
 
			/* use UDP to send it to myself (I use UDP to talk inside TNOS) */
			aprs_send (data_APZxxx, len);
		}

		return 1;
	}

	return 0;
}

#endif

