
/*
 * APRS Services for JNOS 111x, TNOS 2.30 & TNOS 2.4x
 *
 * April,2001-February,2002 - Release # 1.11.2
 *
 * Designed and written by VE4KLM, Maiko Langelaar
 *
 * For non-commercial use (Ham Radio) only !!!
 *
 * 14Nov2001, VE4KLM, MY attempt at tracking and blocking duplicate
 * packets. I need to prevent duplicates in cases like EMAIL server,
 * and other possible services down the road.
 *
 */

#include "global.h"

#ifdef	APRSD

#include "ctype.h"

#ifndef MSDOS
#include <time.h>
#endif

#include "aprs.h"

#define	MAXDUPS	20

typedef struct {

	int16	checksum;
	time_t	last;

} DUPTBL;


static DUPTBL duptbl[MAXDUPS];

static int topcnt = 0;

/*
 * 14Nov2001, Maiko, Taken from smtpserv.c and modified for
 * use in my APRS server code. Compute CCITT CRC-16 of the
 * passed data.  This is 16 bit CRC-CCITT stuff, extracted
 * from Bill Simpson's PPP code.
 */

extern int16 fcstab[];

#define FCS(fcs, c) (((fcs) >> 8) ^ fcstab[((fcs) ^ (c)) & 0x00ff])

static int16 calc_aprs_crc (char *callsign, char *payload)
{
	int16 crc, len;

	crc = 0xffff;

	len = (int16) strlen (callsign);
	while (len--)
		crc = FCS(crc, *callsign++);

	len = (int16) strlen (payload);
	while (len--)
		crc = FCS(crc, *payload++);

	crc ^= 0xffff;

	return crc;
}

int aprs_dup (char *callsign, char *payload)
{
 	int16 checksum = calc_aprs_crc (callsign, payload);
	time_t curtime = time (&curtime);
	int cnt;

	for (cnt = 0; cnt < topcnt; cnt++)
	{
		/* check if this checksum already exists in our dup table */
		if (duptbl[cnt].checksum == checksum)
		{
			/*
			 * make sure it isn't old, if so discard it !!! right now
			 * I consider old to mean anything longer than 5 minutes.
			 */
			if ((curtime - duptbl[cnt].last) < 300)
			{	
				aprslog (-1, "email dup packet");

				return 1;	/* indicate duplicate packet */
			}

			duptbl[cnt].checksum = 0;	/* free the old record for reuse */
		}
	}

	/*
	 * At this point, it's not a dupe, but a first time packet,
	 * so look for an empty slot to put it in if any exist. If
	 * not, put it at the end of the list for now.
	 */

	for (cnt = 0; cnt < topcnt; cnt++)
	{
		/* insert new dup table entry in free slot */
		if (duptbl[cnt].checksum == 0)
		{
			duptbl[cnt].checksum = checksum;	
			duptbl[cnt].last = curtime;

			break;
		}
	}


	/* I should really do this like I did the APRS Heard
	 * table but this is still just an experiment. I just
	 * want some basic protection again duplicate emails
	 * right now, nothing more. This function will get
	 * written better in a subsequent release, it will do
	 * for now. I don't think we will get a full table
	 * on email requests for now :-)
	 */

	if (cnt == topcnt)
	{
		if (topcnt == MAXDUPS)
		{
			aprslog (-1, "warning - can't track any more dupes !!!");
		}
		else
		{
			duptbl[cnt].checksum = checksum;

			duptbl[cnt].last = curtime;

			topcnt++;
		}
	}

	return 0;
}

/* 27Nov2001, Maiko, Some commands to manage packet dups */

static void usagedupcfg (void)
{
	tprintf ("Usage: aprs dup show\n");
}

int doaprsdup (int argc, char *argv[], void *p OPTIONAL)
{
	time_t curtime = time (&curtime);

	int cnt;

	if (argc == 1)
		usagedupcfg ();

	else if (!stricmp (argv[1], "show"))
	{
		tprintf ("APRS Dupe Table (for debugging)\nChecksum  Age\n");

		for (cnt = 0; cnt < topcnt; cnt++)
		{
			tprintf ("%-10.8d %ld\n", duptbl[cnt].checksum,
				curtime - duptbl[cnt].last);
		}
	}

	return 0;
}

#endif

