
/*
 * APRS Services for JNOS 111x, TNOS 2.30 & TNOS 2.4x
 *
 * April,2001-February,2002 - Release # 1.11.2
 *
 * Designed and written by VE4KLM, Maiko Langelaar
 *
 * For non-commercial use (Ham Radio) only !!!
 *
 */

#include "global.h"

#ifdef	APRSD

#define	TNOS241

#include "ctype.h"
#ifdef	JNOSAPRS
#include "cmdparse.h"
#endif
#include "commands.h"

#ifndef MSDOS
#include <time.h>
#endif

#include <sys/stat.h>
#include "mbuf.h"

/* 01Feb2002, Maiko, DOS compile for JNOS will complain without these */
#ifdef  MSDOS
#include "socket.h"
#include "internet.h"
#endif

#include "netuser.h"
#include "ip.h"
#include "files.h"
#include "session.h"
#include "iface.h"
#include "udp.h"
#include "mailbox.h"
#include "usock.h"
#include "mailutil.h"
#include "smtp.h"

#include "aprs.h"	/* 16May2001, VE4KLM, Added prototypes finally */

#if !defined(_lint)
static char rcsid[] OPTIONAL = "$Id: aprssrv.c,v 1.1.1.1 2004/02/09 21:48:03 blantz Exp $";
#endif

#define	ALLOC_BODY	/* This is so that malloc is used to allocate memory
					 * for the body of the HTML 14501 page. I started off
					 * with a static char [] but felt that was wastefull
					 * and not flexible enough. If you don't like the use
					 * of the malloc call, then undefine this. That will
					 * force the use of the static char [], just make sure
					 * it's big enough to hold max aprs heard and the max
					 * number of igate servers in the server list, etc.
					 */

#define	MAXTPRINTF	(SOBUF - 5)

/* 20Jun2001, VE4KLM, The destination call is now defined in aprs.h */
static const char *aprsdestcall = APRSDC;

static int Hsocket = -1;	/* tcpip connection to aprs.net */
static int Asocket = -1;	/* udp listener */

static int monitor_aprsnet = 0;	/* monitoring flag Added 18Apr2001 */

extern char Mycall[AXALEN];	/* TNOS mycall identifier */

static char Aprscall[AXALEN];	/* now use logoncallsign, not Mycall in
								 * our broadcasts and ID, etc */

int aprs_debug = 0;		/* debugging flag */

int aprs_poe = 1;	/* 05Sep2001, VE4KLM, Inject POE call flag */

int aprs_badpackets = 0;	/* 09Jan2002, New flag to keep APRS log small :-) */

/*
 * 22Jan2002, Maiko, APRS IS inactivity timer variables
 */
static struct timer aprs_srv_t;
long aprs_is_timeout = 600;			/* every 10 minutes for now */

/* 25Oct2001, Maiko, APRS specific events now have their own logfile */
static FILE *aprsLogfp = NULLFILE;

/*
 * 10Jul2001, Message Handler definitions
 */

#define	MSG_VIA_RF		1
#define	MSG_VIA_IGATE	0

#define	MAXSRVNAMELEN 20

#define	MAXANETSRV	4

#define	MAXANETCLI	10	/* 02Jan2002, New, 22Jan2002, Bumped up to 10 */

#define IGATE_ONLINE 1

typedef struct {

	char server[MAXSRVNAMELEN];

	int	port;

	int info_port;	/* 11Dec2001, Maiko, New for HTTP link */

	time_t last;	/* 14May2001, Time Connected */

	time_t since;	/* 07Feb2002, Last heard since */

	char flags;

	/* 02Aug2001, VE4KLM, Try incoming Statistics first */
	long in_bytes;
	long in_packets;
	long out_bytes;
	long out_packets;

	long in_packets_last;	/* 25Jan2002, Maiko, Moved this into here, since I'll
							 * need to track client inactivity as well
							 */
} ANETSRV;

/* 07Jan2002, Need more info for client connections */

typedef struct {

	ANETSRV	srv;

	char *user;

	struct timer t;	/* 08Jan2002, timer and socket info */

	int s;

} ANETCLI;

static ANETSRV	anetsrv[MAXANETSRV];
static ANETCLI	anetcli[MAXANETCLI];	/* 02Jan2001, New, 07Jan changed to ANETCLI */

static	int numcli = 0;

static ANETSRV *nxtsrv;

unsigned long internal_hostid = 0;	/* the only kludge for now */

/* 08Nov2001, Maiko, Now a true representation of the email address */
static char *aprs_email = (char*)0;

char logon_callsign[15] = APRSNOCALL;

char aprs_port[15] = APRSNOPORT;

/*
 * 08Nov2001, VE4KLM, New mode for the APRS Email Server
 *
 * 0 means pass through to APRS IS
 * 1 means handle locally and don't pass onto APRS IS
 * 2 means block any EMAIL requests
 *
 */

static int emailmode = 1;	/* default handle locally */

/*
 * APRS Server Commands
 */

/* 31May2001, VE4KLM, hsize for configuring APRS heard table size */
static int dohsize (int argc, char *argv[], void *p);

static int doaprslog (int argc, char *argv[], void *p);
static int doaprserver (int argc, char *argv[], void *p);
static int dosrvcfg (int argc, char *argv[], void *p);
static int dointerface (int argc, char *argv[], void *p);
static int dointernal (int argc, char *argv[], void *p);
static int donetlogon (int argc, char *argv[], void *p);
static int dointmsg (int argc, char *argv[], void *p);
static int dolocmsg (int argc, char *argv[], void *p);

/* 09Jan2002, Replaced dodebug and a few others with doflags */
static int doflags (int argc, char *argv[], void *p);

static int dofwdtorf (int argc, char *argv[], void *p);
static void aprsx (int, void *, void *);
static ANETSRV *next_server (int);

/* 04Aug2001, VE4KLM, Added email for the 14501 page */
static int doemail (int argc, char *argv[], void *p);

/* 08Nov2001, Maiko, New APRS Email Server modes */
static int doemailmode (int argc, char *argv[], void *p);

#ifdef	APRS_14501
/* 22Aug2001, VE4KLM, New function to better alloc body memory */
static int get_srvcnt (void);
#endif

/* 28Nov2001, Maiko, Need a declaration for this function in the
 * smtpserv.c module. Does not seem to be prototyped in any of
 * the standard NOS header files.
 * 04Dec2001, Maiko, In JNOS, this function is declared in mailutil.h
 */
#ifndef	JNOSAPRS
extern int mailuser (FILE *data,const char *from,const char *to,const char *origto);
#endif

/*
 * 27Nov2001, Maiko, Added "BC" and "DUP" subcommands. Moved the
 * bcstat and bcpos into their own module for more flexibility.
 */

#ifdef	JNOSAPRS

static struct cmds DFAR Acmds[] = {

	"bc", doaprsbc, 0, 0, NULLCHAR,
	"dup", doaprsdup, 0, 0, NULLCHAR,
	"flags", doflags, 0, 0, NULLCHAR,
	"fwdtorf", dofwdtorf, 0, 0, NULLCHAR,
	"hsize", dohsize, 0, 0, NULLCHAR,
	"interface", dointerface, 0, 0, NULLCHAR,
	"internal", dointernal, 0, 0, NULLCHAR,
	"listen", doaprserver, 0, 0, NULLCHAR,
	"lmsg", dolocmsg, 0, 0, NULLCHAR,
	"log", doaprslog, 0, 0, NULLCHAR,
	"logon", donetlogon, 0, 0, NULLCHAR,
	"imsg", dointmsg, 0, 0, NULLCHAR,
	"server", dosrvcfg, 0, 0, NULLCHAR,
	"email", doemail, 0, 0, NULLCHAR,
	"emailmode", doemailmode, 0, 0, NULLCHAR,
	NULLCHAR,
};

#else

static struct cmds Acmds[] =
{
	{ "bc", doaprsbc, 0, 0, NULLCHAR },
	{ "dup", doaprsdup, 0, 0, NULLCHAR },
	{ "flags", doflags, 0, 0, NULLCHAR },
	{ "fwdtorf", dofwdtorf, 0, 0, NULLCHAR },
	{ "hsize", dohsize, 0, 0, NULLCHAR },
	{ "interface", dointerface, 0, 0, NULLCHAR },
	{ "internal", dointernal, 0, 0, NULLCHAR },
	{ "listen", doaprserver, 0, 0, NULLCHAR },
	{ "lmsg", dolocmsg, 0, 0, NULLCHAR },
	{ "log", doaprslog, 0, 0, NULLCHAR },
	{ "logon", donetlogon, 0, 0, NULLCHAR },
	{ "imsg", dointmsg, 0, 0, NULLCHAR },
	{ "server", dosrvcfg, 0, 0, NULLCHAR },
	{ "email", doemail, 0, 0, NULLCHAR },
	{ "emailmode", doemailmode, 0, 0, NULLCHAR },
	{ NULLCHAR, NULL, 0, 0, NULLCHAR }
};

#endif

/*
 * 28Nov2001, Maiko, Need some functions to return some of the
 * static variables in this module, needed in aprsbc module. I
 * don't really like doing this (need to structure the globals
 * alot better actually). That can be future cleanup project.
 */

const char *getaprsdestcall (void)
{
	return aprsdestcall;
}

char *getlogoncall (void)
{
	return logon_callsign;
}

int getAsocket (void)
{
	return Asocket;
}

int getHsocket (void)
{
	return Hsocket;
}

/* Send to tnos link level */

int sendrf (char *data, int len)
{
	struct sockaddr_in server_in;

	int s = socket (AF_INET, SOCK_DGRAM, 0);

	server_in.sin_family = AF_INET;
	server_in.sin_port = 16161;

	/*
	 * 30May2001, VE4KLM, If this is not set, then set it to the address
	 * of this TNOS box. This basically means that the 'aprs internal'
	 * autoexec.nos entry is no longer required except in special cases.
	 * ALOT of TNOS systems now have linux doing the encap and tunnel, so
	 * for those systems there is no longer any need for 'aprs internal'.
	 */
	if (internal_hostid == 0)
		internal_hostid = Ip_addr;

	server_in.sin_addr.s_addr = internal_hostid;

	if (sendto (s, data, len, 0, (char*)&server_in, sizeof(server_in)) == -1)
		return -1;

	close_s (s);	/* 15May2001, Oops !!! Not having this is probably the
				 	 * source of my TNOS SIGSEGV problems around sendrf */
	return 0;
}

/*
 * Pointer (to pointer) where we store a callsign list. This pointer
 * is initialized and the callsign list is loaded by the dofwdtorf ()
 * function at the end of this module.
 */
static char **fwdtorf = (char**)0;

/*
 * Function that compares the passed callsign with the entries in
 * our callsign list. If there is a match, then return 1. If there
 * is no match, then return 0.
 *
 * 01Jun2001, VE4KLM, Redesign this so that WILDCARDS can now be
 * used to match callsign patterns. Now sysops can use entries like
 * the following :
 *
 * aprs fwdtorf ve4 k*6 **4 va4
 *
 */

static int callsign_allowed (char *callsign)
{
	char **tptr = fwdtorf;

	char *lptr, *cptr;

	int matched;

	while (*tptr)
	{
		cptr = callsign;
		lptr = *tptr;
		matched = 1;

		while (*lptr)
		{
			if (*lptr != '*' && *lptr != *cptr)
			{
 				matched = 0;
				break;
			}

			cptr++;
			lptr++;
		}

		if (matched)	/* matched !!! return immediately */
			return 1;

		tptr++;		/* next pattern in our list */
	}

	return 0;	/* callsign is not allowed - no match was found */
}

/*
 * 26Oct2001, Maiko, New function to create a directory name and
 * path (if a filename is passed) to the new /nos/spool/aprs area.
 */
static void getaprsdir (char *pname, const char *fname)
{

#ifdef  JNOSAPRS
	sprintf (pname, "%s/aprs", Spoolqdir);
#else
	sprintf (pname, "%s/aprs", Spool);
#endif

	if (fname != NULL)
	{
		strcat (pname, "/");
		strcat (pname, fname);
	}

	/* aprslog (-1, "getaprsdir [%s]", pname); */
}

/* 09Jul2001, Maiko, Log (with time stamp) data to mail/aprs.txt file */
/* 13Jul2001, Maiko, Now have a new APRS directory under Spool area */
/* 16Oct2001, Maiko, Added few more params for more detail */
static void log_aprs_msg (char *ptr, int sending, char *callsign)
{
	char fname[100];
	time_t nowtime;
	FILE *fp;

	/* 13Oct2001, Maiko, Now store APRS messages in the 'Spool/aprs' area */
	/* 26Oct2001, Maiko, New function to do this */
	getaprsdir (fname, "aprs.txt");

	time (&nowtime);

	if ((fp = fopen (fname, "a+")) != NULLFILE)
	{
		fprintf (fp, "%s", ctime (&nowtime));
		fprintf (fp, "%s %s %s\n", callsign, sending ? "sent" : "rcvd", ptr);
		fclose (fp);
	}
}

static void bad_msg (int rf, const char *desc, char *baddata)
{
	if (aprs_badpackets)
	{
		aprslog (-1, "Bad data (%s) from %s - %s",
			desc, rf ? "RF" : "APRS IS", baddata);
	}
}

/*
 * 03Oct2001, Maiko, Separate function to check for a message ID, and
 * then format the proper ack response. I made this a separate function
 * because there are now multiple places in the code where I may be
 * processing incoming messages (ie, the PBBS APRS message center).
 */
static int needmsgack (char *tptr, char *ackmsg)
{
	int len, retval = 0;

	*ackmsg = 0;

	/* search for Message ID identifier */
	while (*tptr && *tptr != '{')
		tptr++;

	/* if one is found, then format the ack message */
	if (*tptr)
	{
		tptr++;	/* skip the identifier */

		/*
		 * make sure we cap the length of the msg id to maximum
		 * of 5 bytes, just incase something is wrong with the data
		 * that we get from the APRS IS - as per the APRS spec.
		 */
		sprintf (ackmsg, "ack%.5s", tptr);

		/* 03Dec2001, Maiko, There could very well be a NEWLINE
		 * character at the end of the incoming message, due to
		 * my code in the aprs.c module, which I should really
		 * change the way I use newlines. Get rid of NEWLINE,
		 * don't let it go out RF in particular.
		 */
		len = strlen (ackmsg);
		if (*(ackmsg+len-1) == '\n')
			*(ackmsg+len-1) = 0;

		retval = 1;
	}

	return retval;
}

/* 05Nov2001, VE4KLM, New email server */

static char DAEMONSTR[] = "%sAPRSSVR@%s (Mail Delivery Subsystem)\n";

static int aprsmail (const char *from, const char *data, const char *pathinfo, int pathinfolen)
{
	char to[50], *toptr = to;
	const char *ptr;
	FILE *tfile;
	int msglen;
	long msgid;
	time_t t;

	/* Extract the email recipient from the APRS data */
	while (*data && *data != ' ')
		*toptr++ = *data++;
	*toptr = 0;

	if (!(*data))
	{
		aprslog (-1, "bad email data [%s]", data);
		return 0;
	}

	data++;	/* skip the space, data now points to body of email */
	
	if ((tfile = tmpfile ()) == NULLFILE)
	{
		aprslog (-1, "can't get a temporary file for email");
		return 0;
	}

	aprslog (-1, "APRS Station %s sending email to %s", from, to);
	aprslog (-1, "Message : %s", data);

	/* count length of the message, excluding msg id at end if one exists */
	for (msglen = 0, ptr = data; *ptr && *ptr != '{'; ptr++, msglen++);

	(void) time (&t);

	fprintf (tfile, "%s%s", Hdrs[DATE], ptime (&t));

	/* 04Dec2001, Maiko, JNOS version has no passed argument */
#ifdef	JNOSAPRS
	msgid = get_msgid ();
#else
	msgid = get_msgid (1);
#endif

	fprintf (tfile, "%s<%ld@%s>\n", Hdrs[MSGID], msgid, Hostname);

	fprintf (tfile, DAEMONSTR, Hdrs[FROM], Hostname);
	fprintf (tfile, "%s%s\n", Hdrs[TO], to);
	fprintf (tfile, "%sMessage from APRS station %s\n\n", Hdrs[SUBJECT], from);

	/* 04Dec2001, Maiko, Put in length parameter so that msd id is not sent
	 * to the receipient of this email. Improves cosmetic appearance
	 */
	fprintf (tfile, "%.*s\n---\n", msglen, data);

	fprintf (tfile, "Please DO NOT reply to this email. The above message was received\n");
	fprintf (tfile, "by the %s Server (IGATE) %s. Inquiries or concerns\n",
		TASVER, logon_callsign);

	fprintf (tfile, "should be directed (by email) to ");

	if (aprs_email)
		fprintf (tfile, "%s.\n", aprs_email);
	else
		fprintf (tfile, "sysop@%s.\n", Hostname);

	fprintf (tfile, "\n%.*s\n", pathinfolen, pathinfo);

	fseek (tfile, 0L, 0);

	/* 04Dec2001, Maiko, JNOS version has one less passed argument */
#ifdef	JNOSAPRS
	(void) mailuser (tfile, "", to);
#else
	(void) mailuser (tfile, "", to, to);
#endif
	(void) fclose (tfile);

	return 1;
}

/*
 * 11Jan2002, Start of new function to better validate
 * the data for a particular DTI. I wrote this mostly
 * because I get periodic beacons in which the first
 * character is often a valid DTI character, but the
 * beacon is not related to APRS at all.
 */

int valid_dti_data (char dti, char *data, int len)
{
	char *bp = data;

	int retval = 1;	/* default to valid */

	switch (dti)
	{
		case 'T':		/* APRS Telemetry Data */

			if (*bp++ != '#')
				retval = 0;
			else if (!isdigit (*bp))
				retval = 0;

			break;

		case '@':

			if (!isdigit (*bp))
				retval = 0;

			break;

	}

	if (!retval && aprs_badpackets)
		aprslog (-1, "DTI %c, Bad Data [%.*s]", dti, len, data);

	return retval;
}

/* 08Jan2002, New function. This one will replace
 * the one in aprs.c (but not quite yet).
 */
int valid_dti (char dti)
{
	int retval = 0;	/* default to invalid */

	switch (dti)
	{
		case 'T':   /* 10Jan2002, Maiko, Oops forgot about APRS telemetry */
		case '@':
		case '!':
		case '#':
		case '$':
		case '%':
		case '*':
		case '/':
		case ':':
		case ';':
		case '<':
		case '=':
		case '>':
		case '?':
		case '_':
			retval = 1;
			break;
	}

	return retval;
}

/* 26Apr2001, Experimental Message Handler */
/* 10Jul2001, Added RF flag so that I can use this function for RF side */
/* 16Jul2001, Igate mode added. If the mode is set to 'manual' then do not
 * send messages from RF to internet, unless the IGATE alias is in the path.
 */
static int message_handler (char *data, int rf)
{
    char callsign[15], *csp = callsign, dti;
	char srccall[15], *ssp = srccall, *ptr;
    int cnt, mymsg, retval, srclen = 0;

	/* Directed Queries and Route Trace support, 29May2001 */
	char tracebuf[100], *savebp = data;
	/* 03Oct2001, Maiko, Bump up mymsgargv size, mods to dointmsg, dolocmsg */
	char *mymsgargv[4], tbuf[300];
	int rfallowed, sendack, dupe;
	void *vp = (void*)0;

	/*
	 * The return value of this function has different meanings depending
	 * on whether the incoming data is from the Internet or local RF port.
	 */
	if (rf)
		retval = 1;
	else
		retval = 0;

	/*
	 * 25May2001, VE4KLM, Unfortunately, we get a fair amount of garbage
	 * from the internet stream. The easiest way to keep alot of it out, is
	 * to check if the data starts with an alphanumeric character. This will
	 * also screen out the REMARKS and COMMENTS that some servers put out.
	 */
	if (!isalnum (*data))
		return retval;

	/* 24May2001, VE4KLM, Get the source callsign from message */
	srclen = 0;
	while (*data && *data != '>')
	{
		/* 16Jul2001, Only need to check for the length of 9 */

		if (srclen > 9)	/* safety feature, incase we get garbage */
		{
			/* 11Jul2001, If error, then show more usefull information */
			/* 25Sep2001, replaced with a common function */
			bad_msg (rf, "Source Callsign", savebp);

			return retval;
		}

		*ssp++ = *data++;

		srclen++;
	}

	*ssp = 0;	/* terminate source callsign string */

	/* 03Oct2001, Maiko, Moved this up here, it's used for more stuff now */
	mymsgargv[1] = srccall;

	/* 25Sep2001, Skip delimiter */
	if (*data == '>')
		data++;
	/*
	 * 13Jul2001, VE4KLM, Ignore calls beginning with APRS because they are
	 * nothing more than server stats. Drop them now, so we don't waste CPU
	 * time, processing a bunch of stuff we'll never use for anything !
	 */
	if (strnicmp (srccall, "aprs", 4) == 0)
		return retval;

	/* skip to end of source path header */
	while (*data && *data != ':')
	{
		/*
		 * 25Sep2001, Maiko, Better checks put in to make sure we don't
		 * get clobbered with bad data from the APRS internet system. This
		 * replaces my old check for soley the '}' characters.
		 *
		 * 26Sep2001, Maiko, Forgot about '/' for dual port tncs. Fixed.
		 *
		 * 14Sep2001, VE4KLM, Put in a check to discard '}' characters
		 * that suddenly appeared in the source path header from the APRS
		 * IS, corrupted packets were coming in, crashing my system.
		 *
		 * 03Dec2001, Maiko, Looks like KH2Z is using '@' for igate POE now.
		 */
		if (!isalnum (*data) && (*data != '-') && (*data != '/') &&
			(*data != '*') && (*data != ',') && (*data != '@'))
		{
			/* 25Sep2001, replaced with a common function */
			bad_msg (rf, "Source Path Header", savebp);
			return retval;
		}

		data++;
	}

	/* just to be on the safe side */
	if (!(*data))
	{
		/* 25Sep2001, replaced with a common function */
		bad_msg (rf, "Truncated Packet", savebp);
		return retval;
	}

	data++;	/* skip the source path header delimiter */

	dti = *data++; /* Get the DTI (Data Type Identifier) */

	/* 08Jan2002, Maiko, Now that I'm accepting client connections, it's
	 * time to make sure we stick with VALID DTI characters as per the latest
	 * specification. I've already noticed one of my clients (APRS+SA) sending
	 * me the old style messages, where there is no DTI character, or the
	 * source path termination character is missing.
	 */
	if (!valid_dti (dti))
	{
		bad_msg (rf, "Bad or missing DTI", savebp);
		return retval;
	}

	/*
	 * 11Jan2002, Maiko, Now make sure the DTI data is at least somewhat
	 * good. New function started to validate stuff particular to the DTI
	 * character in use for this APRS packet.
	 */
	if (!valid_dti_data (dti, data, strlen (data)))
		return retval;

	retval = 0;	/* reset, cause the return value of this function
				 * is different if message comes from IGATE or RF.
				 */

	if (dti == ':') /* Direct or 3rd Party Message */
    {
		/* 29May2001, VE4KLM, Route trace support */
		sprintf (tracebuf, "PATH= %.*s", data - savebp - 2, savebp);

        /* extract the callsign of the 3rd party */
		srclen = 0;
        while (*data && *data != ' ' && *data != ':')
		{
			/* 28Aug2001, Better check the recipient length as well */

			if (srclen > 9)	/* safety feature, incase we get garbage */
			{
				/* 25Sep2001, replaced with a common function */
				bad_msg (rf, "Recipient Callsign", savebp);

				if (rf)
					return (1);
				else
					return (0);
			}

			*csp++ = *data++;

			srclen++;
		}

        *csp = 0;   /* terminate callsign string */
		/*
		 * 23May2001, VE4KLM, Time to start intercepting messages that are
		 * destined for 'us', and handle them accordingly.
	 	 */
		mymsg = 0;

		if (strcmp (callsign, logon_callsign) == 0)
			mymsg = 1;

		/*
		 * 02May2001, VE4KLM, Previously I had hardcoded the VE4 to
		 * be the only callsign allowed (for my testing). Now we use
		 * a callsign list in memory. See dofwdtorf() function.
		 * 10Jul2001, VE4KLM, Don't check for callsign if from RF port.
		 *
		 * 18Oct2001, Maiko, I used to kick it out here if the recipient
		 * callsign was not allowed in the 'fwdtorf' list, but now that I
		 * have the PBBS Message Center, I'll have to move this check after
		 * checking for any active Message Center users. The message center
		 * should be open to ANYONE with permissions, since the message will
		 * never go out to RF anyways if the user is on the PBBS. Only if
		 * the user is not on the PBBS do when then kick them out.
		 */

		rfallowed = 1;
		if (!rf && !mymsg && !callsign_allowed (callsign))
			rfallowed = 0;

        /* skip remaining white space */
        while (*data && *data != ':')
            data++;

        /* skip the message text delimiter */
        data++;

        /* at this point we have the message text */
		/* 24May2001, VE4KLM, Added source call, nice to know who it is */
		/* 18Oct2001, Maiko, Don't show this if call is not allowed on RF */
		if (rfallowed && (aprs_debug || mymsg))
           	aprslog (-1, "[%s] to [%s] %s", srccall, callsign, data);
		/*
		 * 05Nov2001, VE4KLM, Now intercept EMAIL requests, from RF only,
		 * we sure as heck don't want to be processing stuff we get from
		 * the internet side (not yet anyways).
		 * 08Nov2001, Maiko, Added email modes.
	 	 */
		if (emailmode && rf && strcmp (callsign, "EMAIL") == 0)
		{
			if (emailmode == 1)
			{
				aprslog (-1, "email handled locally.");

				/* 20Nov2001, Maiko, Check for duplicate email packet and
				 * only send the email once. IF it's a dupe, it means that
				 * the ack we sent back probably didn't make it, so just
				 * send another one back and see if that one makes it.
			 	 */
				dupe = aprs_dup (srccall, data);

				if (!dupe)
					sendack = aprsmail (srccall, data, savebp, (data - savebp - 2));
				else
					sendack = 1;

				if (sendack)
				{
					/* 09Nov2001, Maiko, Reply with a fake 3rd party message,
					 * this is the only way to make sure the ACK is properly
					 * processed at the originating APRS client. The usual
					 * direct message will not work, since we are not the
					 * station with call EMAIL, we are the station with call
					 * VE4UMR-5. First construct the start of a possible ACK
					 * message, so that if a message id is found, we can
					 * append the ack and complete the message.
					 */
					ptr = tbuf;
    				ptr += sprintf (ptr, "}%s>%s::%-9.9s:", callsign, aprsdestcall, srccall);

					/* 08Nov2001, Maiko, Check for message id, if one is found, then
					 * complete the prebuilt message above, and send it out RF. The only
					 * thing we have to do now is check for duplicate EMAIL requests.
					 */
					if (needmsgack (data, ptr))
					{
						if (sendrf (tbuf, strlen (tbuf)) == -1)
       						aprslog (-1, "sendrf - can't send email ack");
					}

					if (!dupe)
						smtptick (NULL);  /* wake SMTP to send that mail */
				}

				return (1);	/* don't let this go out to the internet */
			}

			if (emailmode == 2)
			{
				aprslog (-1, "email blocked");
				return (1);
			}
		}

		/*
		 * 02Oct2001, VE4KLM, See if the recipient of the message is
		 * currently logged into the PBBS, if so, send it there, and
		 * leave. This code will currently not work for JNOS, sorry.
		 *
		 * 18Oct2001, Maiko, Time to try this on a JNOS system, so
		 * that once again, both TNOS and JNOS will have matching
		 * functionality from the NOSaprs perspective.
		 */

		if (!mymsg)
		{
			/* this is VERY experimental, it's not perfect by a long shot */

			struct mbx *mbxptr = NULLMBX;

#ifdef	JNOSAPRS

        	for (mbxptr = Mbox; mbxptr; mbxptr = mbxptr->next)
			{
			/* name is lower case, callsign is uppercase, use strcasecmp */
				if (!stricmp (mbxptr->name, callsign))
					break;
        	}
#else
			for (cnt = 0; cnt < NUMMBX; cnt++)
			{
				if (Mbox[cnt] != NULLMBX)
				{
				/* name is lower case, callsign is uppercase, use strcasecmp */
					if (!stricmp (Mbox[cnt]->name, callsign))
						break;
				}
			}

			if (cnt != NUMMBX)
				mbxptr = Mbox[cnt];
#endif
			if (mbxptr != NULLMBX)
			{
				/* 03Oct2001, Maiko, Make sure we ack this message
				 * for the user of the PBBS APRS message center.
				 */
				if (needmsgack (data, tbuf))
				{
					mymsgargv[2] = tbuf;

					/* make sure source call is the PBBS user !!! */
					mymsgargv[3] = mbxptr->name;

					/* 10Jul2001, Check if IGATE or RF route */
					if (rf)
						dolocmsg (4, mymsgargv, vp);
					else
						dointmsg (4, mymsgargv, vp);
				}

				/* format the message for PBBS user */
				sprintf (tbuf, "%s: %s\n", srccall, data);
				/* 13Oct2001, Maiko, Log PBBS messages from the internet */
				/* 16Oct2001, Maiko, Added few more params for more detail */
				(void) log_aprs_msg (tbuf, 0, callsign);

				usprintf (mbxptr->user, tbuf);
				usflush (mbxptr->user);

				/* kpause (5000); */

				return retval;	/* message delivered, no point going further */
			}
		}

		/* 18Oct2001, Maiko, We now do this after the PBBS 'see if the user
		 * is using the APRS message center' check above.
		 */
		if (!rfallowed)
			return retval;

		/* At this point, the message is allowed to go out to the RF port,
		 * or it belongs to 'myself'.
		 */
		if (!mymsg)
		{
			if (!rf)
				retval = 1;
		}
		else	
		{
			if (rf)
				retval = 1;

			/* 29May2001, VE4KLM, Respond to some Directed Queries */

			/*
			 * 09Jul2001, Now ack any messages that have a message id. We
			 * should really fork something out to deal with the ack cause
			 * we should send several acks over a period of a few minutes
			 * to ensure the receipient will get it. This is just a test
			 * to see if the following code will work out or not.
			 *
			 * 03Oct2001, Maiko, Put the ack processing in it's own
			 * function (reduces code duplication) because I now have
			 * to do ack processing in several places.
			 */

			if (needmsgack (data, tbuf))
			{
				mymsgargv[2] = tbuf;

				/* 10Jul2001, Check if IGATE or RF route */
				if (rf)
					dolocmsg (3, mymsgargv, vp);
				else
					dointmsg (3, mymsgargv, vp);
			}

			/*
			 * 09Jul2001, VE4KLM, Log messages (with time stamp) to a
			 * file now, as well, send to the console so that the sysop
			 * can finally do some real-time messaging.
			 */

			/* format the console message */
			sprintf (tbuf, "%s: %s\n", srccall, data);
			/* 09Jul2001, Maiko, Log messages from the internet */
			/* 16Oct2001, Maiko, Added few more params for more detail */
			(void) log_aprs_msg (tbuf, 0, callsign);

			/*
			 * 10Oct2001, VE4KLM, No more SYSOP mode for messaging, now
			 * that I have the APRS message console put in. I ran into
			 * complications allowing the APRS Message Center from the
			 * SYSOP mode. I think having the Center from the TNOS console
			 * and the PBBS prompt is sufficient.
			 */
			usprintf (Command->output, tbuf);

			/* Query us for a trace, respond with route trace */
			/* 06Jun2001, VE4KLM, Make this case insensitive */
			if (strnicmp (data, "?PING?", 6) == 0 ||
				strnicmp (data, "?APRST", 6) == 0)
			{
				mymsgargv[2] = tracebuf;

				/* 10Jul2001, Check if IGATE or RF route */
				if (rf)
					dolocmsg (3, mymsgargv, vp);
				else
					dointmsg (3, mymsgargv, vp);
			}
		}
    }

	return retval;
}

/*
 * As of April 11 2000 Steve Dimse has released this following doHash code
 * to the open source aprs community. This allows me to logon to the APRSD
 * system down in Miami, or whoever else is running an APRSD server.
 */

#define kKey 0x73e2	/* This is the key for the data */

static short doHash (const char *theCall)
{
	short hash, i, len;
	char rootCall[10];	/* need to copy call to remove ssid from parse */
	char *ptr, *p1 = rootCall;
	
	while ((*theCall != '-') && (*theCall != 0)) *p1++ = toupper(*theCall++);
	*p1 = 0;
	
	hash = kKey;	/* Initialize with the key value */
	i = 0;
	len = strlen(rootCall);
	ptr = rootCall;
	while (i<len)		/* Loop through the string two bytes at a time */
	{
		hash ^= (*ptr++)<<8;	/* xor high byte with accumulated hash */
		hash ^= (*ptr++);		/* xor low byte with accumulated hash */
		i += 2;
	}

	return hash & 0x7fff; /* mask off high bit so number is always positive */
}

/*
 * The following function is called from config.c, used when parsing
 * the CONSOLE command line or the AUTOEXEC.NOS commands
 */

int doaprs (int argc, char *argv[], void *p)
{
	return subcmd (Acmds, argc, argv, p);
}

/* Connect to World Wide APRS network */

static int connect_aprs_net (ANETSRV *anetsrvp)
{
	struct sockaddr_in fsocket;
	char logonstr[100];
	int comms_socket;

	fsocket.sin_family = AF_INET;

	aprslog (-1, "Connecting Hostname [%s]", anetsrvp->server);

	if ((fsocket.sin_addr.s_addr = resolve (anetsrvp->server)) == 0)
	{
		aprslog (-1, "Not resolvable");
		return -1;
	}

	fsocket.sin_port = anetsrvp->port;

	comms_socket = socket (AF_INET, SOCK_STREAM, 0);

	if (connect (comms_socket, (char*)&fsocket, SOCKSIZE) != 0)
	{
		aprslog (-1, "Connect failed");
		close_s (comms_socket);
		return -1;
	}

	sprintf (logonstr, "user %s pass %u vers %s\n",
		logon_callsign, doHash (logon_callsign), TASVER);

	if (aprs_debug)
		aprslog (-1, "Logon [%s]", logonstr);

	if (send (comms_socket, logonstr, strlen (logonstr), 0) < 1)
	{
		aprslog (-1, "unable to send logon string");
		close_s (comms_socket);
		return -1;
	}

	aprslog (-1, "we are now connected and logged in");

	return comms_socket;
}

/*
 * 01May2001, VE4KLM, Convert a message from the internet to
 * the proper format for delivery over the local RF network
 */

static int convert_3rd_party (unsigned char *buffer)
{
    unsigned char orgmsg[300], *orgptr = orgmsg;

    strcpy (orgmsg, buffer);

    *buffer++ = '}';    /* 3rd Party Data Type Identifier */

    /*
     * copy from incoming source path header until the '*' or if
     * there is no '*', then do the entire source path header, up
     * to the ':' delimiter. Do not include the '*' or ':' chars.
     */
    while (*orgptr && *orgptr != '*' && *orgptr != ':')
        *buffer++ = *orgptr++;

    if (!(*orgptr)) /* just to be safe */
    {
        aprslog (-1, "bad packet, discard");
        return -1;
    }

    /*
     * If there is a '*' at this point we need to skip over it and
     * any other digi callsigns, right up to the ':' character. If
     * there is a ':' at this point, then just skip over it.
     */
    if (*orgptr++ == '*')
    {
        while (*orgptr && *orgptr != ':')   /* skip any digis after '*' */
            orgptr++;

        if (!(*orgptr)) /* just to be safe */
        {
            aprslog (-1, "bad packet, discard");
            return -1;
        }
        orgptr++;   /* skip the ':' delimiter */
    }

    /* the 'orgptr' now points at the APRS message itself ! */

    /*
     * Now insert 3rd party network identifier and gateway callsign,
     * don't forget to prepend the comma, and put '*' at end of 3rd
     * party header, oh yes, and the final ':' character.
     */
    buffer += sprintf (buffer, ",TCPIP,%s*:", logon_callsign);

    /* Finally put in the original APRS message, and we're done */
    strcat (buffer, orgptr);

	return 0;
}

/* 28Nov2001, Maiko, New function to update outgoing stats. Needed
 * because of references in the new aprsbc.o module
 */

void update_outstats (int length)
{
	/* 22Jan2002, Maiko, Carefull, next server may not be available */
	if (nxtsrv == (ANETSRV*)0)
		return;

	/* 02Aug2001, Maiko, Update outgoing stats */
	nxtsrv->out_bytes += length;
	nxtsrv->out_packets++;
}

static void aprsistimeout (void *t OPTIONAL)
{
	static int client_check_counter = 0;

	ANETSRV *aptr;
	ANETCLI *cptr;
	int cnt;

	/*
	 * 22Jan22, Maiko, New function replaces the old reconnect timer which
	 * has been phased out and was not usefull. This one simply checks for
	 * inactivity (ie, a dead link) which happens from time to time with
	 * some of the internet APRS servers. If the number of incoming packets
	 * has not changed since last time, then kill the sockets immediately.
	 */

	if (nxtsrv->in_packets == nxtsrv->in_packets_last)
	{
		aprslog (-1, "IGATE connection appears dead, terminating connection");

		if (Hsocket != -1)
		{
			close_s (Hsocket);
			Hsocket = -1;
		}

		nxtsrv->in_packets_last = -1;	/* make sure we reset this */
	}
	else
		nxtsrv->in_packets_last = nxtsrv->in_packets;
/*
 * 06Feb2002, Maiko, Might as well use this function to check for
 * inactivity on the client connections as well. Active sessions
 * will have the port parameter set to a non-zero value. I should
 * probably put this elsewhere in the future because this function
 * is only active (timed) if there is a connection to the APRS
 * internet system. No big deal for now, it will work here. I
 * use a counter (client_check_counter) so this works out to
 * about once an hour if aprs_is_timeout = 10 minutes.
 */
	if (client_check_counter++ > 5)
	{
		client_check_counter = 0;	/* make sure we reset */

		if (aprs_debug)
			aprslog (-1, "checking for inactive client connections");

		for (cnt = 0; cnt < numcli; cnt++)
		{
			cptr = &anetcli[cnt];

			aptr = &cptr->srv;

			if (aptr->port)
			{
				if (aptr->in_packets == aptr->in_packets_last)
				{
					aprslog (-1, "Client connection appears dead, terminating connection");

					close_s (cptr->s);

					aptr->in_packets_last = -1;	/* make sure we reset this */
				}
				else
					aptr->in_packets_last = aptr->in_packets;
			}
		}
	}

#ifdef	JNOSAPRS
	start_timer (&aprs_srv_t);	/* and restart the timer */
#else
#ifdef	TNOS241
	start_detached_timer (__FILE__, __LINE__, &aprs_srv_t);
#else
	start_detached_timer (&aprs_srv_t);
#endif
#endif

}

/* Read messages coming from ourselves (RF) */

static void aprs_net (int unused OPTIONAL, void *sp, void *p OPTIONAL)
{
	struct sockaddr_in from;
	struct mbuf *bp;
	int fromlen, foo;
	const char *cp;

	struct session *asp = (struct session*)sp;

	while (1)
	{
		fromlen = sizeof (from);

		if ((foo = recv_mbuf (Asocket, &bp, 0, (char *) &from, &fromlen)) == -1)
		{
			if ((cp = sockerr (Asocket)) != NULLCHAR)
				aprslog (-1, "recv (Asocket) - sockerr [%s]", cp);
			else
				aprslog (-1, "recv (Asocket) closed normally");

			break;	/* Server closing */
		}

		if (foo == 0)
			continue;

		if (aprs_debug)
			aprslog (Asocket, "Got RF, %s", bp->data);

		/* 10Jul2001, VE4KLM, Now handle messages from RF side */
		if (!message_handler (bp->data, MSG_VIA_RF))
		{
			if (aprs_debug)
				aprslog (Asocket, "sending to internet");

			/* dont try to send if internet is offline */
			if (Hsocket == -1)
			{
				if (aprs_debug)
					aprslog (-1, "internet offline");

				continue;
			}

			/* send Packet off to the internet APRS server */
			if (send (Hsocket, bp->data, bp->cnt, 0) < 0)
			{
				aprslog (-1, "send internet failed");
				continue;
			}

			/* 22Jan2002, Maiko, Carefull, next server may not be available */
			if (nxtsrv != (ANETSRV*)0)
			{
				/* 02Aug2001, Maiko, Update outgoing stats */
				nxtsrv->out_bytes += bp->cnt;
				nxtsrv->out_packets++;
			}
		}
	}

	asp->proc1 = NULLPROC;

	close_s (Asocket);

	Asocket = -1;
}

/* Process to receive all APRS related messages */

static void aprsx (int unused OPTIONAL, void *u OPTIONAL, void *p OPTIONAL)
{
	unsigned char buffer[300];
	struct sockaddr_in sock;
	struct session sp;
	time_t curtime;
	const char *cp;
	int len;

#ifndef	JNOSAPRS
	server_disconnect_io ();
#endif

	aprslog (-1, "%s startup", TASVER);	/* added 24Aug2001, Maiko */

	/*
	 * 28Nov2001, Maiko, Make sure that some parameters are set
	 * properly. If logon call or the default aprs interface are
	 * not defined at runtime, then abort the server startup !!!
	 */

	if (strcmp (logon_callsign, APRSNOCALL) == 0)
	{
		aprslog (-1, "Abort - APRS Logon callsign not configured !!!");
		return;
	}

	if (strcmp (aprs_port, APRSNOPORT) == 0)
	{
		aprslog (-1, "Abort - APRS interface not configured !!!");
		return;
	}

	/* setup the UDP listener to listen to 'myself' for APRS ax25 frames */

	Asocket = socket (AF_INET, SOCK_DGRAM, 0);
	sock.sin_family = AF_INET;
	sock.sin_addr.s_addr = INADDR_ANY;
	sock.sin_port = 93;

	if (bind (Asocket, (char *) &sock, sizeof (sock)) == -1)
	{
		aprslog (-1, "APRSD: Can't bind");
		Asocket = -1;
		return;
	}

	/* Fork off process to listen to aprs stuff from 'ourselves' - rf */
	sp.proc1 = newproc ("aprs_net", 1024, aprs_net, 0, &sp, NULL, 0);

	/* Start status and position broadcasts to RF side */
	startrfbc ();

	/* Infinite loop, always try to stay connected to an Inet server */

	while (1)
	{
		/* get the next server in this list */
		if ((nxtsrv = next_server (0)) == (ANETSRV*)0)
		{
#ifdef	JNOSAPRS
			pause (30000L);
#else
#ifdef	TNOS241
			(void) kpause (__FILE__, __LINE__, 30000L);	/* 30 seconds */
#else
			(void) kpause (30000L);
#endif
#endif
			continue;
		}

		/* try to connect to it */
		if ((Hsocket = connect_aprs_net (nxtsrv)) == -1)
		{
#ifdef	JNOSAPRS
			pause (30000L);
#else
#ifdef	TNOS241
			(void) kpause (__FILE__, __LINE__, 30000L);	/* 30 seconds */
#else
			(void) kpause (30000L);
#endif
#endif
			continue;
		}

		/* at this point we have a connection, so start tracking connected time */
		nxtsrv->last = time (&curtime);
		nxtsrv->since = nxtsrv->last;	/* 07Feb2002, Maiko, track last heard */

		nxtsrv->flags = IGATE_ONLINE;	/* indicate that it is online (stats) */

		nxtsrv->in_packets_last = -1;	/* 25Jan2002, Maiko, New - RESET for inactivity check */

		nxtsrv->in_bytes = 0;	/* 28Jan2002, Maiko, Reset counters for new connection */
		nxtsrv->in_packets = 0;
		nxtsrv->out_bytes = 0;
		nxtsrv->out_packets = 0;

		/* 22Jan2002, Start inactivity timer for this host connection */

		aprs_srv_t.func = (void (*)(void*)) aprsistimeout; /* call when expired */
		aprs_srv_t.arg = NULL;
		set_timer (&aprs_srv_t, aprs_is_timeout * 1000L); /* set timer duration */

#ifdef	JNOSAPRS
		start_timer (&aprs_srv_t);
#else
#ifdef	TNOS241
		start_detached_timer (__FILE__, __LINE__, &aprs_srv_t);
#else
		start_detached_timer (&aprs_srv_t);
#endif
#endif
		/* 23Jan2002, Maiko, Start status and position broadcasts to INET side */
		startinetbc ();

		/* Now loop till socket get's busted, processing Inet traffic */
		while (1)
		{
			if ((len = recvline (Hsocket, buffer, sizeof(buffer) - 2)) == -1)
			{
				if ((cp = sockerr (Hsocket)) != NULLCHAR)
					aprslog (-1, "recv (Hsocket) - sockerr [%s]", cp);
				else
					aprslog (-1, "recv (Hsocket) closed normally");

				break;	/* Inet Server connection lost or closing */
			}

			if (len == 0)
				continue;

			/* 07Feb2002, Maiko, Now tracking last heard time */
			nxtsrv->since = time (&curtime);

			/* 02Aug2001, Maiko, Update incoming stats */
			nxtsrv->in_bytes += len;	
			nxtsrv->in_packets++;

			/* 19Jul2001, VE4KLM, Get rid of the newline (if there) */
			if (buffer[len-1] == '\n')
				len--;

			buffer[len] = 0;

			/* 16Apr2001, VE4KLM, experimental message handler */
			/* 10Jul2001, VE4KLM, Added flag for RF or IGATE */
			if (message_handler (buffer, MSG_VIA_IGATE))
			{
            	/* 01May2001, VE4KLM, convert message for bc to rf */
            	if (convert_3rd_party (buffer) != -1)
				{
					len = strlen (buffer);	/* MUST recalculate buffer length */

					/* 19Jul2001, VE4KLM, Don't go sending the NULL out RF */
					if (sendrf (buffer, len) == -1)
						aprslog (-1, "sendrf failed");
				}
			}

			if (monitor_aprsnet)
				aprslog (Hsocket, "APRSN: [%.*s]", len, buffer);
		}

		nxtsrv->last = time (&curtime) - nxtsrv->last;	/* end connected time */

		nxtsrv->flags = 0;

		close_s (Hsocket);	/* close the socket and free resources */

		Hsocket = -1;	/* 28Jan2002, Maiko, Oops forgot this one, important */

		stopinetbc (); /* 23Jan2002, Maiko, Stop status and position broadcasts to INET side */

		stop_timer (&aprs_srv_t);	/* make sure we stop inactivity timer */
	}

	/* Stop status and position broadcasts to RF side (will never get here right now) */
	stoprfbc ();
}

#ifdef	APRS_44825

/*
 * 28Dec2001, Maiko, New monitor only port for low bandwidth links or
 * for those stations that simply want to monitor traffic, not handle
 * it, ie, PCSAT listeners.
 * 08Jan2002, Maiko, Added send function (currently only for keep alive
 * status messages since some clients kill connection if they don't
 * receive anything after some time.
 */

static void send44825 (void *ptr)
{
	char obuf[100];
	int len;

	ANETCLI	*cptr = (ANETCLI*)ptr;
	ANETSRV *aptr = &cptr->srv;

	sprintf (obuf, "%s>%s:>keepalive\n", getlogoncall (), getaprsdestcall ());

	len = strlen (obuf) + 1;	/* include string terminator */

	/* 07Feb2002, Maiko, Update client OUT counters */

	if (send (cptr->s, obuf, len, 0) < 0)
	{
		aprslog (-1, "unable to send keepalive");
	}
	else
	{
		aptr->out_bytes += len;
		aptr->out_packets++;
	}

#ifdef  JNOSAPRS
    start_timer (&cptr->t);
#else
#ifdef  TNOS241
    start_detached_timer (__FILE__, __LINE__, &cptr->t);
#else
    start_detached_timer (&cptr->t);
#endif
#endif
}

static void serv44825 (int s, void *unused OPTIONAL, void *p OPTIONAL)
{
	struct sockaddr_in fsocket;
	char outdata[300], *optr;
	char tempuser[20], *tptr;
	time_t curtime;
	ANETSRV *aptr;
	ANETCLI *cptr;
	int len, cnt;

	int user_verified = 0;

	(void) sockmode (s, SOCK_ASCII);
	(void) sockowner (s, Curproc);
	close_s (Curproc->output);
	Curproc->output = s;
	(void) setflush (s, '\n');
	seteol (s, "\r");

	aprslog (s, "connected");

	/* Jan2002, Maiko, Start playing with client connections */

	for (cnt = 0; cnt < numcli; cnt++)
		if (anetcli[cnt].srv.port == 0)
			break;

	if (cnt == numcli)
	{
		if (numcli >= MAXANETCLI)
		{
			aprslog (s, "disconnect, maximum clients");
			close_s (s);
			return;
		}

		numcli++;
	}

	aprslog (s, "Assign IDX %d", cnt);

	cptr = &anetcli[cnt];	/* 07Jan */
	aptr = &cptr->srv;

	if (getpeername (s, (char *) &fsocket, &len) != -1)
		sprintf (aptr->server, "%.19s", inet_ntoa (fsocket.sin_addr.s_addr));
	else
		strcpy (aptr->server, "unknown");

	aptr->last = time (&curtime);
	aptr->since = aptr->last;	/* 07Feb2002, Maiko, Track last heard */

	aptr->port = 14825;
	aptr->info_port = 14501;
	aptr->in_bytes = 0;
	aptr->in_packets = 0;
	aptr->out_bytes = 0;
	aptr->out_packets = 0;

	aptr->in_packets_last = -1;	/* 25Jan2002, Maiko, RESET inactivity check */

	/* 02Jan2002, VE4KLM, end of monitor code */

	cptr->s = s;	/* important for timer function */

	cptr->t.state = 0;	/* probably a good idea */

	while (1)
	{
		if ((len = recvline (s,outdata, sizeof(outdata) - 2)) == -1)
			break;

		if (aprs_debug)
		{
			aprslog (s, "MON B1 (%d) B2 (%d) L1 (%d) L2 (%d)",
				*outdata, *(outdata+1), *(outdata+len-1), *(outdata+len-2));

			aprslog (s, "len %d [%s]", len, outdata);
		}

		/* Why not just write it out the Hsocket ? Well, I should
		 * really validate the incoming data before I simply go shooting
		 * it out to the APRS internet system. Using aprs_send () will
		 * forward this packet to the main APRS server which contains
		 * the data validation routines, etc.
		 */

		if (len > 1)
		{
			/* 07Feb2002, Maiko, Track since last heard */
			aptr->since = time (&curtime);

			/* update incoming stats */
			aptr->in_bytes += len;
			aptr->in_packets++;

			/* 07Jan2002, Maiko, User info and authentication */
			if (!user_verified)
			{
				optr = outdata;

				/* seems we get leading CR from UI-View logon string */
				if (*optr == '\r' || *optr == '\n')
					optr++;

				if (!memcmp (optr, "user", 4) || !memcmp (optr, "USER", 4))
				{
					optr = skipwhite (optr + 4);

					tptr = tempuser;

					while (*optr && *optr != ' ')
						*tptr++ = *optr++;

					*tptr = 0;	/* terminate string */

					aprslog (s, "User [%s] logging in", tempuser);

					cptr->user = strdup (tempuser);

					/* find out which software version they are using */
					tptr = strstr (optr, "vers");

					if (*tptr == 0)
						tptr = strstr (optr, "VERS");

					if (*tptr)
						tptr = skipwhite (tptr + 4);

					if (*tptr == 0)
					{
						aprslog (s, "software not specified");
						break;
					}

					aprslog (s, "Software [%s]", tptr);

					/* User is verified and connected for sure only after
					 * we confirm that some type of software is specified.
					 */
					user_verified = 1;

					/* WinAPRS and APRS+SA so far will kill the connection if they
					 * don't receive anything after a few minutes or so. UI-View so
					 * far does not care if it doesn't get anything. If this is not
					 * a UI-View system, then start the keepalive broadcasts.
					 */
					if (memcmp (tptr, "UI", 2) == 0)
						continue;

					/*
					 * keepalive is experimental, only do for APRS+ stations at this
					 * time since no others have complained about disconnecting after
					 * a few minutes of inactivity. Search for 'vers' identifier
					 * 11Jan2002, VE4KLM, Reduce from 120 to 90, that may be cutting
					 * it fine, as Walter (K5WH) is still getting disconnects from
					 * time to time on his APRS+SA connection to my port 14825.
					 */
					aprslog (s, "keepalive timer enabled");

					/* 08Jan2002, Maiko, Keep alive timer for connection */
					cptr->t.func = (void (*)(void*)) send44825;
					cptr->t.arg = (void*)cptr;
					set_timer (&cptr->t, 90000L);

					/* send a keep alive right now to get timer going */
					send44825 ((void*)cptr);

					continue;
				}
				else
				{
					aprslog (s, "No user authentication");

					break;
				}
			}

			len++;	/* include NULL terminator */

			/* for piggy backed messages, that have a NULL terminator
			 * on the first message (only noticed with my software so
			 * far) need to look at this sending of the NULL somemore
			 */
			if (*outdata == 0 || *outdata == '\n' || *outdata == '\r')
				aprs_send (outdata + 1, len - 1);
			else
				aprs_send (outdata, len);
		}
	}

	aprslog (s, "Free IDX %d", cnt);

	aptr->port = 0;	/* indicate a free slot */

	stop_timer (&cptr->t);	/* 08Jan2002 */

	aprslog (s, "disconnected");

	close_s (s);
}

/*
 * 28Dec2001, Maiko, Functions to stop and start the monitor port
 */

static const char *mondescstr = "44825";

static int S44825 = -1;

int t44825on (int argc, char *argv[], void *p OPTIONAL)
{
#ifdef	JNOSAPRS
		S44825 = 1;
		return (start_tcp (14825, mondescstr, serv44825, 512));
#else
		argc = 1;	/* Use the passed port number 14825 below */

		return (installserver (argc, argv, &S44825, mondescstr, 14825,
			INADDR_ANY, mondescstr, serv44825, 512, NULL));
#endif
}

int t44825off (int argc OPTIONAL, char *argv[] OPTIONAL, void *p OPTIONAL)
{

#ifdef	JNOSAPRS
	S44825 = -1;
	stop_tcp (14825);
#else
	deleteserver (&S44825);
#endif

	return 0;
}

#endif		/* end of APRS_44825 */

#ifdef	APRS_14501

/* 07Jan2002, Maiko, Modified to include user info if Server Connection banner, I have one
 * function to handle the banner for Host and User connections (saves code). Both are almost
 * the same (were, until I added more info for the Client connections).
 */

static char *dsp14501hdr (const char *contype, char *ptr)
{
	/* show igate connection status list */
	ptr += sprintf (ptr, "<h4>%s</h4><table border=1 cellpadding=3 bgcolor=\"#00ffdd\"><tr>", contype);

	if (*contype == 'C')
		ptr += sprintf (ptr, "<td>User</td>");

	/* 07Feb2002, Maiko, Added 'Since last heard' header column */
	ptr += sprintf (ptr, "<td>Server</td><td>Port</td><td>Connected</td><td>Pkts In</td><td>Bytes</td><td>Pkts Out</td><td>Bytes</td><td>Since</td></tr>");

	return ptr;
}

/*
 * 31Jul2001, VE4KLM, New 14501 Information Server, so finally we can go
 * direct to aprsstat from any of the APRSD 14501 info pages on the web.
 */

static void serv14501 (int s, void *unused OPTIONAL, void *p OPTIONAL)
{

#ifndef	ALLOC_BODY
	static char body[2000];	/* this 14501 sucker can take alot of memory :-( */
	char *ptr = body;
#else
	char *body, *ptr;
#endif

	char *tptr, tindstr[25];
	time_t curtime, nowtime;
	int needed, len, cnt;
	struct mbuf *bp;

	int hdrflg = 0;	/* only display IGATE Con header if servers are present */

	int dspdone = 0;	/* 14Nov2001, Maiko, Wierd behavior on some servers */

	(void) sockmode (s, SOCK_BINARY);
	(void) sockowner (s, Curproc);
	close_s (Curproc->output);
	Curproc->output = s;
	(void) setflush (s, '\n');

	aprslog (s, "connected");

	while (recv_mbuf (s, &bp, 0, NULLCHAR, NULL) > 0)
	{
		aprslog (s, "14501 page");

		/* 14Nov2001,Maiko, I don't think this is a good idea anymore,
		 * especially considering it's purely for HTML browsers. Also,
		 * I've noticed that this text based stuff is being appended
		 * to some browser output on certain platforms running NOSaprs,
		 * makes me wonder if we are getting zero length data here, so
	 	 * I think I'll check for that as well,hopefully that solves
		 * the appearance of the text based stuff for good. Also put
		 * in a fail safe maybe, that if the HTML has already been
		 * 'displayed', then don't display it again. From what I understand,
		 * it's the browsers job to terminate the connection, so we don't
		 * break out, we just continue with the recv_mbuf () call.
		 */

		if (bp->cnt == 0 || dspdone)
		{
			aprslog (-1, "DEBUG : zero length msg or done already ? Flag : %d", dspdone);
			continue;
		}

		free_p (bp);

		dspdone = 1;	/* flag it as done, so we don't generate
					 	 * multiple pages if that is possible */
#ifdef	ALLOC_BODY

		/* 22Aug2001, Maiko, now actually check # of servers in list,
		 * and we now track digi (so we need additional 60 or so).
		 * 11Dec2001, Added 30 and 30 for new server host info link
		 * 02Jan2001, Maiko, Added allocation for client connect.
		 */ 
		needed = 1400 + get_srvcnt() * 190 + numcli * 190 + get_topcnt () * 180;

		aprslog (-1, "numcli %d alloc %d bytes for body space", numcli, needed);

		body = malloc (needed);

/* this is better approach. I don't use the
	mallocw - there is absolutely no reason
	why TNOS should shutdown simply cause
	there's no memory to do a 14501 page. */

		if (body == (char*)0)
		{
			aprslog (s, "No memory available");
			break;
		}

		ptr = body;
#endif
		/* put general information first */

		/* 11Dec2001, Maiko, Moved TNOS version up here instead of at the end */
		ptr += sprintf (ptr, "<html><center><br><b>APRS Server & Internet Gateway - VE4KLM-%s</b><br><font size=2>powered by KA9Q NOS - %s</font><br><br><table border=1 cellpadding=3 bgcolor=\"#00ffee\"><tr><td>Machine</td><td>Uptime</td><td>Callsign</td><td>Interface</td><td>Contact Information</td></tr><tr><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>", TASVER, Version, Hostname, tformat (secclock ()), logon_callsign, aprs_port);

		/* 08Nov2001, Maiko, I now need to have the email address as a string elsewhere, so I moved the href
		 * code into here now. Previously aprs_email was a string that contained the HTML formatting, now it
		 * is not, since I need the actually raw email address somewhere else in this code (aprsmail).
		 */
		if (aprs_email)
			ptr += sprintf (ptr, "<a href=\"mailto:%s\">%s</a>", aprs_email, aprs_email);
		else
			ptr += sprintf (ptr, "<a href=\"mailto:sysop@%s\">sysop@%s</a>", Hostname, Hostname);

		ptr += sprintf (ptr, "</td></tr></table>");

		for (hdrflg = cnt = 0; cnt < MAXANETSRV; cnt++)
		{
			if (anetsrv[cnt].port)
			{
				nowtime = time (&curtime);

				if (anetsrv[cnt].flags == IGATE_ONLINE)
				{
					/* 23Aug2001, Maiko, Now shade the row, instead of using
					 * an asterick symbol like was done for the text version
					 */
					tptr = tformat ((int32)(nowtime - anetsrv[cnt].last));

					strcpy (tindstr, "<tr bgcolor=\"#ffff00\">");
				}
				else
				{
					/* 08Feb2002, If never connected before, then indicate it */
					if (anetsrv[cnt].last)
						tptr = tformat ((int32)anetsrv[cnt].last);
					else
						tptr = (char*)"-";

					strcpy (tindstr, "<tr>");
				}

				/* 29Aug2001, Maiko, Make this like the text based on, if no
				 * servers are configured, then don't bother putting up this
				 * header. Instead, indicate later that no servers present.
				 */
				if (!hdrflg)
				{
					ptr = dsp14501hdr ("Server Connection", ptr);
					hdrflg = 1;
				}

				/* 11Dec2001, Maiko, Added configurable link for host information */
				ptr += sprintf (ptr, "%s<td><a href=\"http://%s:%d\">%s</a></td>", tindstr,
						anetsrv[cnt].server, anetsrv[cnt].info_port, anetsrv[cnt].server);

				ptr += sprintf (ptr, "<td>%d</td><td>%s</td><td>%ld</td><td>%ld</td><td>%ld</td><td>%ld</td>",
						anetsrv[cnt].port, tptr, anetsrv[cnt].in_packets, anetsrv[cnt].in_bytes,
							anetsrv[cnt].out_packets, anetsrv[cnt].out_bytes);

				/* 08Feb2002, If never connected before, then indicate it */
				if (anetsrv[cnt].since)
						tptr = tformat ((int32)(nowtime - anetsrv[cnt].since));
				else
						tptr = (char*)"-";

				/* 07Feb2002, Maiko, Now show 'since last heard' info */
				ptr += sprintf (ptr, "<td>%s</td></tr>", tptr);
			}
		}

		/* 29Aug2001, VE4KLM, Make sure we complete table if header was
		 * actually inserted. If not, then just say that no IGATES are
		 * configured, just like we do for the text based version.
		 */
		if (hdrflg)
			ptr += sprintf (ptr, "</table><br>");
		else
			ptr += sprintf (ptr, "<h4>Internet APRS server(s) not configured</h4>");

		/* 02Jan2002, Maiko, Added Client connections to the monitoring */
		for (hdrflg = cnt = 0; cnt < numcli; cnt++)
		{
			if (anetcli[cnt].srv.port)
			{
				nowtime = time (&curtime);

				tptr = tformat ((int32)(nowtime - anetcli[cnt].srv.last));

				if (!hdrflg)
				{
					ptr = dsp14501hdr ("Client Connections", ptr);
					hdrflg = 1;
				}

				/* ptr += sprintf (ptr, "<tr bgcolor=\"#ffff00\"><td>  "); */

				/* 07Jan2002, Maiko, Client links will not have HTMP link, add user logon for client */
				ptr += sprintf (ptr, "<tr><td>%s</td><td>%s</td>", anetcli[cnt].user, anetcli[cnt].srv.server);

				ptr += sprintf (ptr, "<td>%d</td><td>%s</td><td>%ld</td><td>%ld</td><td>%ld</td><td>%ld</td>",
						anetcli[cnt].srv.port, tptr, anetcli[cnt].srv.in_packets, anetcli[cnt].srv.in_bytes,
							anetcli[cnt].srv.out_packets, anetcli[cnt].srv.out_bytes);
				/* 07Feb2002, Maiko, Now show 'since last heard' info */
				ptr += sprintf (ptr, "<td>%s</td></tr>",
						tformat ((int32)(nowtime - anetcli[cnt].srv.since)));
			}
		}

		if (hdrflg)
			ptr += sprintf (ptr, "</table><br>");
		else
			ptr += sprintf (ptr, "<h4>No APRS clients connected at this time</h4>");

		/* put in the users heard information */
		ptr += build_14501_aprshrd (ptr);

		/* end the page (20Nov2001, Maiko, Added Version of TNOS itself) */
		/* 11Dec2001, Maiko, Moved version to the top instead */
		ptr += sprintf (ptr, "</center></html>\r\n");

		len = ptr - body;

		aprslog (-1, "Body length %d", len);

		/* write the HEADER record */
		/* 03Dec2001, Maiko, Change to 1.1 from 1.0 */
		/* 04Dec2001, Maiko, Change back to 1.0, take out close cmd */
		tprintf ("HTTP/1.0 200 OK\r\n");

		/* 03Dec2001, Maiko, Force client to close session when done
		tprintf ("Connection: close\r\n");
		*/
		tprintf ("Content-Length: %d\r\n", len);
		tprintf ("Content-Type: text/html\r\n");
		tprintf ("Server: %s\r\n", TASVER);

		/* use a BLANK line to separate the BODY from the HEADER */
		tprintf ("\r\n");

		/* 01Aug2001, VE4KLM, Arrggg !!! - the tprintf call uses vsprintf
		 * which has a maximum buf size of SOBUF, which itself can vary
		 * depending on whether CONVERSE or POPSERVERS are defined. If
		 * we exceed SOBUF, TNOS is forcefully restarted !!! Sooooo...,
		 * we will have to tprintf the body in chunks of SOBUF or less.
		 *
		tprintf ("%s", body);
		 */

		ptr = body;

		while (len > 0)
		{
			if (len < MAXTPRINTF)
			{
				tprintf ("%s", ptr);
				len = 0;
			}
			else
			{
				tprintf ("%.*s", MAXTPRINTF, ptr);
				len -= MAXTPRINTF;
				ptr += MAXTPRINTF;
			}
		}

#ifdef	ALLOC_BODY
		/* aprslog (s, "free up body space"); */
		free (body);
#endif
		break;	/* FORCE termination of connection */
	}

	aprslog (s, "disconnected");

	close_s (s);
}

/*
 * 24Sep2001, Maiko, Time to do this right. Removed the 'aprs 14501 on'
 * method and the 'do14501' function. This server is now started and stopped
 * using the standard 'dostart', and 'dostop' functions in config.c.
 */

static const char *descstr = "14501";

static int S14501 = -1;

int t14501on (int argc, char *argv[], void *p OPTIONAL)
{
#ifdef	JNOSAPRS
		S14501 = 1;
		return (start_tcp (14501, descstr, serv14501, 512));
#else
		argc = 1;	/* Use the passed port number 14501 below */

		return (installserver (argc, argv, &S14501, descstr, 14501,
			INADDR_ANY, descstr, serv14501, 512, NULL));
#endif
}

int t14501off (int argc OPTIONAL, char *argv[] OPTIONAL, void *p OPTIONAL)
{

#ifdef	JNOSAPRS
	S14501 = -1;
	stop_tcp (14501);
#else
	deleteserver (&S14501);
#endif

	return 0;
}

#endif	/* end of APRS_14501 */

/* Start or Stop  the APRS daemon */

static int doaprserver (int argc, char *argv[], void *p OPTIONAL)
{
	if (argc == 1)
		tprintf ("APRS Server: O%s\n", (Asocket != -1) ? "n" : "ff");

	else
	{
		if (!stricmp (argv[1], "on"))
		{
			if (Asocket == -1)
			{
				/* Start domain server task */
				(void) newproc ("APRS Server", 1024, aprsx, 0, NULL, NULL, 0);
			}
		}
		else
		{
			if (Asocket != -1)
			{
				close_s (Asocket);
				Asocket = -1;
			}

			if (Hsocket != -1)
			{
				close_s (Hsocket);
				Hsocket = -1;
			}
		}
	}

	return 0;
}

/* 31May2001, VE4KLM, Configure size of (and init) APRS heard table */

static int dohsize (int argc, char *argv[], void *p OPTIONAL)
{
	static int hsize = 0;

	if (argc == 1)
	{
		tprintf ("aprs hsize %d\n", hsize);
	}
	else
	{
		/* 28Aug2001, VE4KLM, The sysop should have the ability to disable
		 * the heard table at any time (by setting hsize to 0).
		 */ 
		if (*argv[1] == '0')
		{
			aprslog (-1, "aprs heard table disabled");
			hsize = 0;
		}
		else if ((hsize = atoi (argv[1])) < 1)
		{
			tprintf ("Bad APRS hsize value");
			return -1;
		}

		init_aprshrd (hsize);
	}

	return 0;
}

/*
 * 09Jan2002, Maiko, New single function for general flags. The idea
 * is to do away with each function dedicated to a single flag. Might
 * as well have it all done by one single function. This replaces the
 * dodebug, domonitor, and dopoe functions. New badpackets flag.
 */

static int doflags (int argc, char *argv[], void *p OPTIONAL)
{
	char *ptr;
	int state;

	if (argc == 1)
	{
		tprintf ("aprs flags : %cdebug %cbadpackets %cmonitor %cpoe\n",
			aprs_debug ? '+' : '-', aprs_badpackets ? '+' : '-',
			monitor_aprsnet ? '+' : '-', aprs_poe ? '+' : '-');
	}
	else
	{
		while (argc > 1)
		{
			ptr = argv[argc-1];

			state = 0;

			if (*ptr++ == '+')
				state = 1;

			if (!stricmp (ptr, "debug"))
				aprs_debug = state;

			if (!stricmp (ptr, "badpackets"))
				aprs_badpackets = state;

			if (!stricmp (ptr, "monitor"))
				monitor_aprsnet = state;

			if (!stricmp (ptr, "poe"))
				aprs_poe = state;

			argc--;
		}
	}

	return 0;
}

/* Return the next available APRSNET server */

static ANETSRV *next_server (int beginning)
{
	static int current_server = 0;

	int cnt;

	if (beginning)
		current_server = 0;

	for (cnt = current_server; cnt < MAXANETSRV; cnt++)
	{
		if (anetsrv[cnt].port)
		{
			current_server = cnt + 1;	/* point to the next one */
			return &anetsrv[cnt];
		}
	}

	if (cnt >= MAXANETSRV)
		current_server = 0;

	return (ANETSRV*)0;
}

#ifdef	APRS_14501

/*
 * 22Aug2001, Maiko, Get the number of servers in our list, so that
 * I can better allocate the amount of memory needed for the 14501
 * body statistics
 */
static int get_srvcnt (void)
{
	int cnt = 0, count = 0;

	while (cnt < MAXANETSRV)
	{
		if (anetsrv[cnt].port)
			count++;

		cnt++;
	}

	/* aprslog (-1, "%d servers in the list", count); */

	return count;
}

#endif

/* Add or Delete from list of APRSNET servers */

static void usagesrvcfg (void)
{
	tprintf ("Usage: aprs server add|delete|list|kick *[parameters]\n");
}

static int dosrvcfg (int argc, char *argv[], void *p OPTIONAL)
{
	int cnt, port;

	if (argc == 1)
		usagesrvcfg ();

	else if (!stricmp (argv[1], "add"))
	{
		/* 24May2001, VE4KLM, Check number of args, or it could crash */
		/* 13Dec2001, Maiko, Possibility of 5 args now, make it less than */
		if (argc < 4)
		{
			usagesrvcfg ();
			return 0;
		}

		for (cnt = 0; cnt < MAXANETSRV; cnt++)
		{
			if (!anetsrv[cnt].port)
			{
				if ((port = atoi (argv[3])) < 1)
				{
					tprintf ("invalid port\n"); 
					break;
				}
				if (strlen (argv[2]) > MAXSRVNAMELEN)
				{
					tprintf ("server name is too long\n"); 
					break;
				}

				strcpy (anetsrv[cnt].server, argv[2]);
				anetsrv[cnt].port = port;

				/* 11Dec2001, Maiko, New info port for HTTP info link */
				anetsrv[cnt].info_port = 14501;

				/* 13Dec2001, Maiko, Sysop can now set the info port to what they want */
				if (argc == 5)
				{
					if ((port = atoi (argv[4])) < 1)
						tprintf ("invalid info port, defaulting to 14501\n"); 
					else
						anetsrv[cnt].info_port = port;
				}

				/* 02Aug2001, VE4KLM, Init statistics counters */
				anetsrv[cnt].in_bytes = 0;
				anetsrv[cnt].in_packets = 0;

				break;
			}
		}

		if (cnt == MAXANETSRV)
			tprintf ("no more room to add additional servers\n");
	}
	else if (!stricmp (argv[1], "delete"))
	{
		/* 24May2001, VE4KLM, Check number of args, or it could crash */
		if (argc != 3)
		{
			usagesrvcfg ();
			return 0;
		}

		for (cnt = 0; cnt < MAXANETSRV; cnt++)
		{
			if (strcmp (anetsrv[cnt].server, argv[2]) == 0)
			{
				anetsrv[cnt].port = 0;
				break;
			}
		}

		if (cnt == MAXANETSRV)
			tprintf ("could find the requested server for deletion\n");
	}
	else if (!stricmp (argv[1], "list"))
	{
		tprintf ("\nCurrent APRS.NET Servers configured\n");

		for (cnt = 0; cnt < MAXANETSRV; cnt++)
		{
			if (anetsrv[cnt].port)
			{
				tprintf ("%-20.20s %5d\n", anetsrv[cnt].server, anetsrv[cnt].port);
			}
		}
	}
	/*
	 * 07Feb2002, Maiko, Added a kick option so we can force rotation to next
	 * server in the list. Wish I had put this in months ago, that way if the
	 * current APRS internet system server is acting up, we can bounce them,
	 * and keep TNOS running (no restart or reordering of server list).
	 */
	else if (!stricmp (argv[1], "kick"))
	{
		if (Hsocket != -1)
		{
			close_s (Hsocket);
			Hsocket = -1;
		}
	}

	return 0;
}

static int dointernal (int argc, char *argv[], void *p OPTIONAL)
{
	if (argc < 2)
	{
		tprintf ("aprs server internal hostid %s\n", inet_ntoa (internal_hostid));
		return 0;
	}

	if ((internal_hostid = resolve (argv[1])) == 0)
	{
		tprintf ("internal hostid not resolvable\n"); 
		return -1;
	}

	aprslog (-1, "internal hostid %s", inet_ntoa (internal_hostid));

	return 0;
}

/* 08Nov2001, VE4KLM, New email server modes */
static int doemailmode (int argc, char *argv[], void *p OPTIONAL)
{
	if (argc == 2)
	{
		if (stricmp ("passthru", argv[1]) == 0)
			emailmode = 0;
		else if (stricmp ("local", argv[1]) == 0)
			emailmode = 1;
		else
			emailmode = 2;
	}

	return 0;
}

/* 04Aug2001, VE4KLM, Added email parameter for 14501 page */
static int doemail (int argc, char *argv[], void *p OPTIONAL)
{
	if (argc == 2)
	{
		aprslog (-1, "aprs email %s", argv[1]);

		/* 08Nov2001, Maiko, I now need to store the raw email address,
		 * previously I had it so that aprs_email string contained the
		 * email address formated into HREF within HTML, no more. Also,
		 * no more length restrictions anymore for this (ie, used to be
		 * capped at 50 bytes, no more).
		 */
		aprs_email = strdup (argv[1]);
	}

	return 0;
}

static int donetlogon (int argc, char *argv[], void *p OPTIONAL)
{
	char pname[100];

	if (argc < 2)
	{
		tprintf ("aprs logon callsign %s\n", logon_callsign);
		return 0;
	}

	if (strlen (argv[1]) > 10)
	{
		/* I should probably validate this better */
		tprintf ("callsign is too long\n");
		return 0;
	}

	strcpy (logon_callsign, argv[1]);

	strupr (logon_callsign); /* 06Jun2001, VE4KLM, Force it uppercase */

	/* 17May2001, VE4KLM, Now use logon as the actual 'mycall' */
	if (setcall (Aprscall, logon_callsign) == -1)
	{
		aprslog (-1, "setcall failed [%s], using 'mycall'", logon_callsign);

		/* default to TNOS 'mycall' variable */
		memcpy (Aprscall, Mycall, AXALEN);

		return 0;
	}

	aprslog (-1, "aprs logon callsign [%s]", logon_callsign);

	/* 26Oct2001, Maiko, Make sure the APRS subdirectory exists
	 * under the /nos/spool area. I was doing this previously in
	 * the log_aprs_msg () function, but really we only need to
	 * do it only once (not each time a message is logged), so
	 * I moved it into here along with a new function.
	 */

	getaprsdir (pname, NULL);

/* 01Feb2002, Maiko, DOS compile for JNOS does not like mkdir */
#ifndef	MSDOS
	(void) mkdir (pname, 0777);
#endif

	return 0;
}

/* 09Jul2001, This one let's me send messages to users on local RF port */

static int dolocmsg (int argc, char *argv[], void *p OPTIONAL)
{
    char msgbuf[200];

    if (argc < 3)
    {
        tprintf ("usage : aprs lmsg <callsign> <message>\n");
        return 0;
    }

	(void) strupr (argv[1]);	/* 19Jul2001, VE4KLM, Force uppercase */

    sprintf (msgbuf, ":%-9.9s:%s", argv[1], argv[2]);

	if (aprs_debug)
    	aprslog (-1, "sending msg [%s]", msgbuf);

	if (sendrf (msgbuf, strlen (msgbuf)) == -1)
       	aprslog (-1, "sendrf - can't send local message\n");

	/* 09Jul2001, Maiko, Log messages to the internet */
	sprintf (msgbuf, "%s: %s\n", argv[1], argv[2]);
	/* 16Oct2001, Maiko, Added few more params for more detail */
	(void) log_aprs_msg (msgbuf, 1, logon_callsign);

    return 0;
}

/* Internal test function that lets me send messages over the internet */

static int dointmsg (int argc, char *argv[], void *p OPTIONAL)
{
    char msgbuf[200], srccall[20], *srccall_cp;

    if (argc < 3)
    {
        tprintf ("usage : aprs msg <callsign> <message>\n");
        return 0;
    }

	/*
	 * 03Oct2001, Maiko, Need a way to override the src call. I never
	 * anticipated this originally, but I need it now for the new APRS
	 * message center available at the PBBS prompt. This is for internal
	 * use only, and should be hidden from the usage phrase above.
	 */
	if (argc == 4)
	{
		strcpy (srccall, argv[3]);
		(void) strupr (srccall);	/* force uppercase */
		srccall_cp = srccall;
	}
	else
		srccall_cp = logon_callsign;

	(void) strupr (argv[1]);	/* 19Jul2001, VE4KLM, Force uppercase */

    sprintf (msgbuf, "%s>%s::%-9.9s:%s\n",
		srccall_cp, aprsdestcall, argv[1], argv[2]);

	if (aprs_debug)
    	aprslog (-1, "sending msg [%s]", msgbuf);

    if (send (Hsocket, msgbuf, strlen (msgbuf) + 1, 0) < 0)
        tprintf ("unable to send message\n");

	/* 22Jan2002, Maiko, Carefull, next server may not be available */
	if (nxtsrv != (ANETSRV*)0)
	{
		/* 02Aug2001, Maiko, Update outgoing stats */
		nxtsrv->out_bytes += (strlen (msgbuf) + 1);
		nxtsrv->out_packets++;
	}

	/* 09Jul2001, Maiko, Log messages to the internet */
	sprintf (msgbuf, "%s: %s\n", argv[1], argv[2]);
	/* 16Oct2001, Maiko, Added few more params for more detail */
	(void) log_aprs_msg (msgbuf, 1, srccall_cp);

    return 0;
}

static int dointerface (int argc, char *argv[], void *p OPTIONAL)
{
    if (argc < 2)
    {
        tprintf ("aprs interface (port) is '%s'\n", aprs_port);
        return 0;
    }

    if (strlen (argv[1]) > 10)
    {
        tprintf ("too long, use shorter name\n");
        return -1;
    }

    strcpy (aprs_port, argv[1]);

    aprslog (-1, "aprs interface (port) is '%s'\n", aprs_port);

    return 0;
}

/*
 * Load into memory a list of callsigns and/or callsign patterns that we
 * will allow to be forwarded from the internet to the local RF port.
 *
 * This function uses a somewhat elegant method to create a DYNAMICALLY
 * allocated list of strings in memory. Nothing is presumed, memory is
 * allocated as needed to store a simple list of strings. Using pointers
 * like this is a good way to conserve memory.
 *
 * May 2, 2001, VE4KLM, Pointers to pointers are the way to go :-)
 *
 * Use the function, callsign_allowed (char*) to verify whether or not
 * a particular callsign is in the callsign list.
 *
 */

static int dofwdtorf (int argc, char *argv[], void *p OPTIONAL)
{
	char **tptr;

	int cnt;

    if (argc < 2)
    {
		/* show it to us if already loaded into memory */
		if (fwdtorf)
		{
			tptr = fwdtorf;

			tprintf ("These callsigns are allowed to go out the RF port\n");

			while (*tptr)
			{
				tprintf ("%s\n", *tptr);

				tptr++;
			}
		}
		else
		{
			/* otherwise display usage */
			tprintf ("aprs fwdtorf [callsigns &| patterns]\n");
		}

		return 0;
	}

	/*
	 * If the callsign list is already allocated, then tell the user
	 * about it. I should really change this so that the user can delete
	 * the list, then recreate it again. This will do for now. It's only
	 * a prototype at this stage, still have some ideas floating around.
	 *
	 * 01Jun2001, VE4KLM, Can now reconfigure this on the fly !
	 */
	if (fwdtorf)
	{
		/* If callsign list is already present, delete it first */
		tptr = fwdtorf;

		/* first free all the strdups originally allocated */
		while (*tptr)
		{
			free (*tptr);
			tptr++;
		}

		/* now free the list of pointers to the strdups */
		free (fwdtorf);

		fwdtorf = (char**)0;
	}

	/* Build Callsign List */

	if ((fwdtorf = (char**)malloc (sizeof(char*) * (argc+2))) == (char**)0)
	{
		tprintf ("not enough memory for callsign list\n");
		return -1;
	}

	/* now load the arguments into the memory list */
	for (tptr = fwdtorf, cnt = 0; cnt < (argc - 1); cnt++)
	{
		*tptr = strdup (argv[cnt+1]);

		strupr (*tptr); /* 06Jun2001, VE4KLM, force upper case */

		tptr++;
	}

	*tptr = (char*)0;	/* terminate the list with NULL */

	return 0;
}

/*
 * The following 2 functions used to be in the file, aprstorf.c, when I
 * originally released my prototypes. That file is no longer present, all
 * that code is now merged into this source file (aprssrv.c). I did it to
 * say in sync with what Brian Lantz did when he incorporated the APRS stuff
 * into his latest release of TNOS 2.40 on May 14, 2001.
 *
 * aprsd_frame ()
 * aprsd_torf ()
 *
 */

int aprsd_frame (udp)
	struct udp *udp;
{
    return (udp->dest == 16161);
}

void aprsd_torf (iface, bp, ip)
      struct iface *iface;
      struct mbuf *bp;
      struct ip *ip OPTIONAL;
{
	struct mbuf *hbp, *data;
	char recipient[AXALEN];
	struct ax_route *axr;
	char dest[AXALEN];
	char const *idest;
	struct ax25 addr;
	char *rptr, *ptr;
	char rtmp[15];

	const char *hrdifacename = NULL;	/* important initialization !!! */

	/* 24Oct2001, Maiko, I used to get the iface parameters for the
	 * main APRS port at this point, but now that I'm checking the
	 * heard table, there's no point looping through the iface list
	 * twice. So the code has been moved further down after we find
	 * out who the destination is (recipient or aprsdest call).
	 */

	ptr = bp->data;

	/* 17May2001, VE4KLM, I want to use the sendrf function for more
	 * than just 3rd party messaging now. I want to broadcast a status
	 * and position for this IGATE server, and possibly other stuff down
	 * the road - so now we have to look at the first character of the
	 * message to distinguish whether it is 3rd party or not.
	 * 10Jul2001, VE4KLM, Sysop can now send local aprs messages to users
	 * on the RF port, so we'll have to route those by recipient as well,
	 * meaning we'll have to add the check for the ':' DTI below.
	 */

	if (*ptr == '}' || *ptr == ':')
	{
		/*
		 * Extract the recipient from the APRS text message, because
		 * we need to setup routing based on the recipient. I know, I
		 * know, that goes against the concepts of WIDE, RELAY, etc,
		 * but I designed this TNOS implementation for a bit more
		 * flexibility. Why should people 3 netrom nodes away not
		 * be able to participate in the APRS network ???
		 */
	
		if (*ptr == '}')	/* 11Jul2001, only do this for 3rd party */
		{	
			/* look for message datatype id */
			while (*ptr != ':')
				ptr++;

			ptr++;	/* skip it, check next one */
		}

		if (*ptr != ':')
		{
			aprslog (-1, "no delimiter, descarding");
			return;
		}

		ptr++;	/* skip it */

		/* extract the message recipient */
		rptr = rtmp;
		while (*ptr && *ptr != ' ' && *ptr != ':')
			*rptr++ = *ptr++;

		/* terminate the recipient string */
		*rptr = 0;

	}
	else
	{
		/* non 3rd party frames are always destined for 'APZ000' */
		strcpy (rtmp, aprsdestcall);
	}

	if (setcall (recipient, rtmp) == -1)
	{
		aprslog (-1, "setcall failed [%s]", rtmp);
		return;
	}

	if (setcall (dest, aprsdestcall) == -1)
	{
		aprslog (-1, "setcall failed [%s]", aprsdestcall);
		return;
	}

	/* 24Oct2001, Maiko, New way to process the interface parameters. Only
	 * one loop is required. This new system incorporates the APRS heard table
	 * so that we should technically now be able to handle APRS messaging on
	 * ANY of the TNOS ax.25 interfaces, not just the default one anymore.
	 */

    if (strcmp (rtmp, aprsdestcall))
	{
		hrdifacename = iface_hrd_on (recipient);
		if (hrdifacename && aprs_debug)
		{
			aprslog (-1, "recipient [%s] last heard on iface [%s]",
				rtmp, hrdifacename);
		}
	}

	/* if not in heard table, or if aprsdestcall, then use the default port */
	if (hrdifacename == NULL)
		hrdifacename = aprs_port;

	for (iface = Ifaces; iface != NULLIF; iface = iface->next)
	{
		if (strcmp (iface->name, hrdifacename) == 0)
			break;
	}

	if (iface == NULLIF)
	{
        aprslog (-1, "No APRS interface available !!!");
		return;
	}

	/*
	 * If there's a digipeater route, get it. Please note, I originally
	 * wanted to use the TNOS 'axsend' function here to save code space,
	 * but the digipeater route needs to be setup based on the dest call
	 * of the message receipient, NOT the APRS dest call of APZ000. So I
	 * have to essentially write my own version of axsend to do this.
	 *
	axsend (iface, dest, Mycall, LAPB_COMMAND, UI, data);
	 */

#ifdef	JNOSAPRS
	axr = ax_lookup (recipient, iface);
#else
	axr = ax_lookup (Aprscall, recipient, iface, AX_NONSETUP);
#endif

	/*
	 * 05Jun2001, VE4KLM, OOPS !!! Even though I had wanted 3rd party msgs
	 * to be routed based on recipient, we still need to be in compliance
	 * with APRS routing concepts. If there is no route defined for the
	 * recipient, then route based on the aprs dest call. Why ? The nice
	 * thing about NOS is that you can use the command, 'ax25 route add',
	 * to configure the APRS unproto path (nice eh ???). For example :
	 * 
	 *    ax25 route add APZ000 ax1 WIDE WIDE'
	 */

    if (strcmp (rtmp, aprsdestcall) && axr == NULLAXR)
	{
#ifdef	JNOSAPRS
		axr = ax_lookup (dest, iface);
#else
		axr = ax_lookup (Aprscall, dest, iface, AX_NONSETUP);
#endif
	}

	memcpy (addr.dest, dest, AXALEN);
	memcpy (addr.source, Aprscall, AXALEN);

	addr.cmdrsp = LAPB_COMMAND;

    if (axr != NULLAXR)
	{
        memcpy (addr.digis, axr->digis, (size_t)axr->ndigis * AXALEN);
        addr.ndigis = axr->ndigis;
        idest = addr.digis[0];
    }
	else
	{
        addr.ndigis = 0;
        idest = dest;
    }

    addr.nextdigi = 0;

	/* allocate memory for new message (include 2 bytes for CTL & PID) */
    if ((hbp = alloc_mbuf ((int16)(bp->cnt + 2))) == NULLBUF)
	{
		aprslog (-1, "no memory for new message");
	    return;
	}

	/* set control and pid fields */
	hbp->data[0] = uchar (UI);
	hbp->data[1] = uchar (PID_NO_L3);

	/* copy the APRS data to buffer */
    hbp->cnt = (int16)(bp->cnt + 2);
    memcpy (hbp->data + 2, bp->data, bp->cnt);

	/* now put in the ax25 header */
    if ((data = htonax25 (&addr, hbp)) == NULLBUF)
	{
		aprslog (-1, "could not put in the ax25 header");
        free_p (hbp);
        return;
    }

	if (aprs_debug)
	{
    	aprslog (-1, "sending recipient [%s] port [%s] %d digis",
			rtmp, aprs_port, addr.ndigis);
	}

	/* send the packet out !!! */
	(*iface->raw)(iface, data);

	return;
}

static int dspigatehdr (void)
{
	tprintf ("Server connection:\n     Server            Port  Connected   Pkts In   Bytes     Pkts Out  Bytes\n");

	return 1;
}

static int dspclihdr (void)
{
	tprintf ("Client connections:\n  User       Server       Port  Connected   Pkts  Bytes In   Pkts  Bytes Out\n");

	return 1;
}

/* 11May2001, VE4KLM, Fingerd statistics so we can put on our website */

int doaprsstat (int argc OPTIONAL, char *argv[] OPTIONAL, void *p OPTIONAL)
{
	char tind, *tptr;
	register int cnt;
	time_t curtime;
	int hdrflg = 0;

	/*
	 * 17May2001, VE4KLM, Added the version information. It could be
	 * very useful down the road, since I am still doing development
	 * on this stuff and probably will be for the next little while.
	 * 20Jun2001, VE4KLM, Added uptime.
	 * 31Jul2001, VE4KLM, Changed the look again.
	 */
	tprintf ("Ver: %s, Uptime: %s, Call: %s, Interface: %s\n\n",
		TASVER, tformat (secclock ()), logon_callsign, aprs_port);

	for (hdrflg = cnt = 0; cnt < MAXANETSRV; cnt++)
	{
		if (anetsrv[cnt].port)
		{
			if (anetsrv[cnt].flags == IGATE_ONLINE)
			{
				tptr = tformat ((int32)(time (&curtime) - anetsrv[cnt].last));
				tind = '*';
			}
			else
			{
				tptr = tformat ((int32)anetsrv[cnt].last);
				tind = ' ';
			}

			if (!hdrflg)
				hdrflg = dspigatehdr ();

			/* display server name and port */
			tprintf ("%c %-20.20s %-5d %s  %-9ld %-9ld %-9ld %ld\n",
				tind, anetsrv[cnt].server, anetsrv[cnt].port, tptr,
					anetsrv[cnt].in_packets, anetsrv[cnt].in_bytes,
						anetsrv[cnt].out_packets, anetsrv[cnt].out_bytes);
		}
	}

	/* 20Jun2001, VE4KLM, If no IGATE servers are configured, then say so */
	if (!hdrflg)
		tprintf ("Internet APRS server(s) not configured\n");

	tprintf ("\n");

	for (hdrflg = cnt = 0; cnt < MAXANETCLI; cnt++)
	{
		if (anetcli[cnt].srv.port)
		{
			tptr = tformat ((int32)(time (&curtime) - anetcli[cnt].srv.last));

			if (!hdrflg)
				hdrflg = dspclihdr ();

			/* display client name and port */
			tprintf ("%-10.10s%-16.16s%-6d%s  %-6ld%-11ld%-6ld%ld\n", anetcli[cnt].user,
				anetcli[cnt].srv.server, anetcli[cnt].srv.port, tptr, anetcli[cnt].srv.in_packets,
					anetcli[cnt].srv.in_bytes, anetcli[cnt].srv.out_packets, anetcli[cnt].srv.out_bytes);
		}
	}

	if (!hdrflg)
		tprintf ("No APRS clients connected at this time\n");

    /* 14May2001, VE4KLM, More user stats */
    display_aprshrd ();

	return 0;
}

/*
 * NEW for start of October. The start of an APRS message center for the
 * general PBBS user, this is under development and is VERY experimental,
 * so don't expect anything real fancy yet - I'm just playing around.
 * 10Oct2001, Can also be used from the console now.
 * 18Oct2001, Maiko, Time to try this with a JNOS system, should work
 * fine now that I've figured out a few things.
 */
int dombaprs (int argc OPTIONAL, char *argv[] OPTIONAL, void *p)
{
	static long msgidpool = 1L;	/* very simple msgid pool for starters */
	struct mbx *m = (struct mbx *) p;
	char tmpbuf[20], srccall[20], recipient[20], cmd_c;
	const char *huhstr = "Huh ?\n", *pingstr = "?APRST";
	char *mymsgargv[4], *cp;
	void *vp = (void*)0;
	int usemsgid = 1;	/* 03Oct2001, Auto message ID for outgoing messages */
	/* 10Oct2001, VE4KLM, Now useable from the console */
	int isconsole = (Curproc->input == Command->input);
	struct mbuf *bp = NULLBUF;	/* MUST do this, or else memory fault */

	/* 10Oct2001, Maiko, Permission not required on TNOS console */
	if (!isconsole)
	{
		/* 11Oct2001, Maiko, Do not allow this from sysop mode, it get's
		 * too complicated trying to redirect the acks back there. Having
		 * the APRS message center available from the PBBS prompt and the
		 * TNOS console should be more than enough for now.
		 *
		 * 13Oct2001, The MBX state is not bitmapped, use '==', not '&'.
		 *
		 * 17Oct2001, Moved this conditional up here, and added a check
		 * to see if the 'm' structure pointer is NULL. Funny why this
		 * would only show up in TNOS 2.41. If someone tries to access
		 * the APRS message center from sysop mode, it seems that 'm' is
		 * set to NULL (normal actually), but why is it not happening in
		 * my TNOS 2.30 release ? Anyways, bottom line, need to check for
		 * 'm' pointer value == NULL, or else it faults on us ...
		 */
		if (m == NULL || m->state == MBX_SYSOP)
		{
			tputs ("Sorry, can only be used from PBBS or TNOS console.\n");
			return 0;
		}

		/* 02Oct2001, Maiko, Must have telnet permission (security) */
		if (!(m->privs & TELNET_CMD))
		{
			tputs (Noperm);
			return 0;
		}

	}

	/* 10Oct2001, VE4KLM, For TNOS console, use the logon_callsign */
	if (!isconsole)
	{
		strcpy (srccall, m->name);
		(void) strupr (srccall); /* Force uppercase */

	/* 11Oct2001, Maiko, Another snag. If the PBBS user callsign is the
	 * same as the LOGON_CALLSIGN, then tell the user to use the TNOS console
	 * instead (assumption made that it is probably the sysop of the TNOS
	 * system himself (or herself) trying to do this. That's reasonable.
	 */
		if (strcmp (srccall, logon_callsign) == 0)
		{
			tputs ("System callsign detected. Please use the TNOS console.\n");
			return 0;
		}

		/* 25Oct2001, Maiko, Wouldn't hurt to log usage */
		aprslog (-1, "%s using APRS msg center", m->name);
	}
	else
		strcpy (srccall, logon_callsign);

	/* 02Oct2001, Maiko, If our connection to APRS IS is down, no point */
	if (Hsocket == -1)
	{
		tputs ("The Internet System is OFFLINE - please try again later.\n");
		return 0;
	}

	tprintf ("Welcome to the %s Message Center. Use '/h' for help.\n", TASVER);
	tputs ("Don't forget to set the msg recipient - use '/r callsign'.\n");

	*recipient = 0;

	while (1)
	{
		usflush (Curproc->output);

		/* 10Oct2001, The PBBS prompt uses the mbxrecvline () call, while
		 * the TNOS console uses the recv_mbuf on the Command input stream.
		 */
		if (!isconsole)
		{
			if (mbxrecvline (m) < 0)
				break;

			cp = m->line;
		}
		else
		{
			free_p (bp);

			if (recv_mbuf (Command->input, &bp, 0, NULLCHAR, 0) < 0)
				break;

			cp = bp->data;

			rip (cp);	/* need to do this or we get trailing junk */
		}

#ifdef MBXTDISC
		/* Only if not on TNOS console */
		if (!isconsole)
			stop_timer (&m->tdisc);
#endif
		/* no point doing anything if nothing was entered */
		if (*cp == '\0')
		{
			tputs (huhstr);
			continue;
		}

		/* Now parse out special message center commands */
		if (*cp == '/')
		{
			cp++;

			cmd_c = *cp++;

			if (cmd_c == 'q')
			{
				tputs ("73!\n");
				break;
			}
			else if (cmd_c == 'r' && *cp == ' ')
			{
				sprintf (recipient, "%.9s", skipwhite (cp));
				(void) strupr (recipient); /* Force uppercase */
		/*
		 * 12Oct2001, Maiko, Validate the AX25 callsign of the recipient. So
		 * far this is really the only function that I can find in TNOS to do
		 * this. It's not perfect, but it's better than nothing.
		 */
				if (setcall (tmpbuf, recipient) == -1)
				{
					tputs ("Invalid callsign, try it again.\n");
					*recipient = 0;
				}
				else
					tprintf ("Msg recipient is now %s.\n", recipient);
			}
			else if (cmd_c == 'p')
			{
				cp = (char*)pingstr;
			}
			else if (cmd_c == 'm')
			{
				tputs ("Msg ids are now on.\n");
				usemsgid = 1;
			}
			else if (cmd_c == 'n')
			{
				tputs ("Msg ids are now off.\n");
				usemsgid = 0;
			}
			else if (cmd_c == 'h')
			{
				tputs ("/m - attach msg id to msgs - expect acks.");
				tputs ("\n/n - do not use msg ids - no acks.");
				tputs ("\n/p - send an APRS ping to current recipient.");
				tputs ("\n/q - quit and return to PBBS prompt.");
				tputs ("\n/r callsign - set new msg recipient.\n");
			}	
			else tputs (huhstr);

			/* we want the ping command to go through */
			if (cmd_c != 'p')
				continue;
		}

		/* Recipient MUST be defined if we want to send out a message */
		if (*recipient == 0)
		{
			tputs ("You didn't set the msg recipient - use '/r callsign'.\n");
			continue;
		}

		/* 03Oct2001, Concatonate a msg ID to end of message if enabled */
		/* Oops, don't do this for query type messages (APRS Spec says no) */
		if (usemsgid && *cp != '?')
		{
			tprintf ("* sending with msgid %ld *\n", msgidpool);

			sprintf (tmpbuf, "{%ld", msgidpool++);		
			strcat (cp, tmpbuf);
			/* Can't let this go over 5 digits, as if it will ever happen */
			if (msgidpool > 99999L)
				msgidpool = 0;
		}

		/* 03Oct2001, Use existing functions to send to the APRS IS */
		mymsgargv[1] = recipient;
		mymsgargv[2] = cp;
		mymsgargv[3] = srccall;

		dointmsg (4, mymsgargv, vp);

#ifdef MBXTDISC
		/* Only if not on TNOS console */
		if (!isconsole)
		{
#ifdef	TNOS241
			start_detached_timer (__FILE__, __LINE__, &m->tdisc);
#else
			start_detached_timer (&m->tdisc);
#endif
		}
#endif
	}

	if (isconsole)
		free_p (bp);

	return 0;
}

/*
 * 25Oct2001, Maiko, Time to put APRS logging to it's own
 * file. I've had several people *complain* about this, and
 * rightfully so. I shouldn't be using the main log, it just
 * happened that way (it was convenient at the time) The two
 * functions below are adaptations of the main logging functions
 * that are defined in the TNOS 'main.c' source file.
 */

int doaprslog (int argc, char *argv[], void *p OPTIONAL)
{
	static char *aprslogname = NULLCHAR;

	char fname[128];

	if (argc < 2)
	{
		if (aprsLogfp && aprslogname != NULLCHAR)
			tnoslog (-1, "APRS Logging to %s", aprslogname);
		else
			tnoslog (-1, "APRS Logging off");

		return 0;
	}

	if (aprsLogfp)
	{
		tnoslog (-1, "APRS log closed");
		(void) fclose (aprsLogfp);
		aprsLogfp = NULLFILE;
		free (aprslogname);
		aprslogname = NULLCHAR;
	}

	if (strcmp (argv[1], "stop") && strcmp (argv[1], "off"))
	{
		strncpy (fname, make_fname (Command->curdirs->dir, argv[1]), 128);
		aprslogname = strdup (fname);
		aprsLogfp = fopen (aprslogname, APPEND_TEXT);
	}

	return 0;
}

#ifdef LOG_GMT_TZ
#define TZTIME(x) asctime(gmtime(x))
#else
#define TZTIME(x) ctime(x)
#endif

void aprslog (int s, const char *fmt,...)
{
	va_list ap;
	char *cp;
	time_t t;
	int i;
	struct sockaddr fsocket;
	char buf[SOBUF];

#ifdef	MSDOS
	int fd;
#endif

	if (aprsLogfp == NULLFILE)
		return;

	va_start (ap, fmt);
	(void) vsprintf (buf, fmt, ap);
	va_end (ap);

	(void) time (&t);
	cp = TZTIME (&t);
	rip (cp);
	i = SOCKSIZE;
	fprintf (aprsLogfp, "%s", cp);
	if (getpeername (s, (char *) &fsocket, &i) != -1)
		fprintf (aprsLogfp, " %s", psocket (&fsocket));

	fprintf (aprsLogfp, " - %s\n", buf);
	(void) fflush (aprsLogfp);

#ifdef	MSDOS
	/* MS-DOS doesn't really flush files until they're closed */
	fd = fileno (aprsLogfp);
	if ((fd = dup (fd)) != -1)
		close (fd);
#endif

}

#endif
