#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#undef TNOS_MALLOC
#include "dirutil.h"

#undef fopen
#undef fclose

char *WhitePages = (char *) "spool/wpagebbs";
#define LINELEN 128
#define	READ_TEXT	"r"
#define	APPEND_TEXT	"a+"


void wpageUpdate (FILE *fp, char *entry);
char * wpageCheck (char *string);
void wpageAdd (char *entry);
#ifndef mallocw
void *mallocw(unsigned size);

void j_free (const void *p)
{
    free (p);
}

void *
mallocw(unsigned size) {
    return (malloc (size));
}
#endif


void
wpageUpdate(FILE *fp, char *entry)
{
time_t now;

	time(&now);
	fprintf(fp,"%-32s %-14ld\n",entry,now); // update time stamp
}


/* Returns a copy (strdup'ed) of existing entry if string is in whitepages file */
char *
wpageCheck(char *string)
{
char buf[LINELEN], *cp, *retval = NULLCHAR;
FILE *fp;

	if (string && ((fp = fopen(WhitePages,"r+t")) != NULLFILE))	{
		while(fgets(buf, LINELEN, fp) != NULLCHAR)	{
			/* Get rid of following spaces or tabs, to separate the
			 * h_addr and it's time stamp entry
			 */
			if( ((cp=strchr(buf,' ')) != NULLCHAR) || \
				((cp=strchr(buf,'\t')) != NULLCHAR) )
					*cp = '\0';
			if(!strnicmp(string,buf, strlen(string)))	{
				retval = strdup (buf);
				fseek (fp, (long) -49, SEEK_CUR);
				wpageUpdate (fp, buf);
				break;
			}
		}
		fclose(fp);
	}
	return retval;
}
     

void
wpageAdd (char *entry)
{
time_t now;
FILE *fp;
char *last;

	if (entry == NULLCHAR)
		return;
	last = wpageCheck (entry);
	free (last);
	if (!last)	{
		time(&now);
		if((fp = fopen(WhitePages,APPEND_TEXT)) != NULLFILE) {
			fprintf(fp,"%s %-14ld\n",entry,now); /* Save h_addr in whitepages file */
			fclose(fp);
		}
	}

}


int main (int argc OPTIONAL, char *argv[] OPTIONAL)
{
FILE *tmp_file;
struct ffblk ff;
char buf[LINELEN], *cp, *cp2;
char fname[32];

	strcpy(buf,"spool/mail/*.txt");
	if (findfirst(buf, &ff, 0) == 0) {
		do	{
			sprintf (fname, "spool/mail/%s", ff.ff_name);
			if ((tmp_file = fopen (fname,READ_TEXT)) != NULLFILE)	{
				while(fgets(buf,LINELEN,tmp_file) != NULLCHAR)	{
					if(!strnicmp(buf,"R:",2)) { /*found one*/
						/* Find the '@[:]CALL.STATE.COUNTRY'or
						 * or the '?[:]CALL.STATE.COUNTRY' string
						 * The : is optional.
						 */
						if( ((cp=strchr(buf,'@')) != NULLCHAR) ||
							((cp=strchr(buf,'?')) != NULLCHAR) ) {
								if((cp2=strchr(cp,' ')) != NULLCHAR)
									*cp2 = '\0';
								if((cp2=strchr(cp,'\n')) != NULLCHAR)
									*cp2 = '\0';
								if((cp2=strchr(cp,'\t')) != NULLCHAR)
									*cp2 = '\0';
								/* Some bbs's send @bbs instead of @:bbs*/
								if (*++cp == ':')
									cp++;
							    	wpageAdd (cp);
							    	printf("%s: Adding '%s'\n", ff.ff_name, cp);
							}
					}
				}
				fclose (tmp_file);
			}

		} while (findnext(&ff) == 0);
	}
    return 0;
}



