#include "global.h"
#include "ctype.h"
#include "color.h"
#include "socket.h"
#include "session.h"
#define COLOR_BLACK	0
#define COLOR_RED	1
#define COLOR_GREEN	2
#define COLOR_YELLOW	3
#define COLOR_BLUE	4
#define COLOR_MAGENTA	5
#define COLOR_CYAN	6
#define COLOR_WHITE	7

#if !defined(_lint)
static char rcsid[] OPTIONAL = "$Id: color.c,v 1.1.1.1 2004/02/09 21:49:07 blantz Exp $";
#endif

#ifdef DEMO
#define tprintf printf
#define tputc putchar

static unsigned char last[2];

void
tflush ()
{
	fflush (stdout);
}
#else
#include "usock.h"

static void myusflush (int s);

static void
myusflush (s)
int s;
{
register struct usock *up;

	if((up = itop(s)) == NULLUSOCK)
		return;

	if(up->obuf == NULLBUF)
		return;
	if (len_p (up->obuf) < 220)
		return;
	usflush (s);
}
#endif

static int getmycolor (unsigned char color);
static int getone (FILE *fp);
static void getkeyword (unsigned char *input,FILE *fp);
static int cmdmatch (unsigned char *input);
static int blinking (unsigned char c);
static int bold (unsigned char c);
int getcolor (char *str);
int docolor (int argc, char *argv[], void *p);
void setscreens (int back, int fore, int clr);
extern void displayStatLine (int offset, int phase, int onlymarquee);


int SYSback = COLOR_BLUE, SYSfore = COLOR_WHITE;
#ifdef SM_CURSES
extern short STATback, STATfore;	/* statline colors */
extern short SPLITfore, SPLITback;	/* chatline colors */
#endif

extern int init_pairs(void);
extern int STATLINE;

static char nope[] = "Color '%s' unknown!\nAvailable colors: black, red, green, cyan, blue, magenta, and white\n";



int
colorprintf (char *escapeflag, char usecolor, const unsigned char *str)
{
unsigned char c;
int foundone = 0;


	while ((c = *str++) != 0)	{
		if (c == 0x1b)	{	/* escape sequence */
			if (usecolor)	{
				foundone = 1;
#ifdef DEMO
				tflush ();
#else
				myusflush(Curproc->output);
#endif
			}
			if (escapeflag)
				*escapeflag = 1;
		}
		else if (escapeflag && *escapeflag)	{
			if (!isdigit(c) && c != ';' && c != '[')
				*escapeflag = 0;
		}
		if (!escapeflag || (escapeflag && !*escapeflag) || usecolor)	{
			tputc (c);
		}
	}
	return (foundone);
}


static int
getmycolor (unsigned char color)
{
int retval = 30;

	switch (color)	{
		case '1': case '9':	retval += 4;	break;
		case '2': case 'A':	retval += 2;	break;
		case '3': case 'B':	retval += 6;	break;
		case '4': case 'C':	retval += 1;	break;
		case '5': case 'D':	retval += 5;	break;
		case '6': case 'E':	retval += 3;	break;
		case '7': case 'F':	retval += 7;	break;
		default:		break;
	}
	return (retval);
}


static int
getone (fp)
FILE *fp;
{
int c;

	c = getc (fp);
	if (feof (fp))
		c = 0;
	return (c);
}


static void
getkeyword (input, fp)
unsigned char *input;
FILE *fp;
{
unsigned char c = 0;
int k;

	do	{
		for (k = 0; k < 15; k++)	{
			input[k] = c = uchar(getone (fp));
			if (!c || (c == '@'))
				break;
		}
		input[k] = 0;
		if (!c)
			k++;
	} while (!k);
}


void
colorcls ()
{
#ifdef DEMO
	tflush ();
#else
	myusflush(Curproc->output);
#endif
	tprintf ("\033[2J");
}


static int
cmdmatch (input)
unsigned char *input;
{
int retval = 0;

	if (!strcmp ("CLS", (char *) input))	{
		colorcls();
		retval = 1;
	}
	return (retval);
}


static int
blinking (unsigned char c)
{
	return ((c > '7') ? 1 : 0);
}


static int
bold (unsigned char c)
{
	return ((c > '7') ? 1 : 0);
}


int
colorchange (input, last)
register const char *input;
register char *last;
{
int retval = 0, putone = 0, temp, putall = 0;


	if (!input || !last)
		return (retval);
	if (!strncmp (input, last, 2))
		return (retval);
	if ((strlen(input) == 2) && (isxdigit (input[0])) && (isxdigit (input[1])))	{
#ifdef DEMO
		tflush ();
#else
		myusflush(Curproc->output);
#endif
		tprintf ("\033[");
		retval = 2;
		if (!*last)
			putall++;
		if (putall || (blinking(uchar(*last)) && !blinking(uchar(*input))) || (bold (uchar(last[1])) && !bold(uchar(input[1])))) {
			tputc ('0');
			retval++;
			putone++;
			putall++;
		}
		if (blinking(uchar(*input)))	{
			tprintf ("%s5", putone ? ";" : "");
			putone++;
			retval++;
		}
		temp = getmycolor(uchar(*input));
		if (putall || temp != getmycolor (uchar(*last)))		{
			tprintf ("%s%d", putone ? ";" : "", temp + 10);
			retval += (2 + putone);
			putone++;
		}
		if (putall || (input[1] != last[1]))	{
			tprintf ("%s%s%d", putone ? ";" : "", (input[1] > '7') ? "1;" : "",
				getmycolor (uchar(input[1])));
			retval += (putone + 2 + ((input[1] > 7) ? 2 : 0));
		}
		tputc ('m');
		retval += 1;
		memcpy (last, input, 2);
	}
	return (retval);
}


void
colorfile (filenm, last)
const char *filenm;
char *last;
{
FILE *fp;
unsigned char c, input[16];
char invalid = 0;

	if ((fp = fopen (filenm, "rb")) == NULLFILE)
		return;
	while (!feof (fp))	{
		if ((c = (unsigned char) getone (fp)) == '\0')
			continue;
		if (c == 0x1b)
			invalid = 1;
		if (c == '@')	{
			invalid = 0;
			getkeyword (input, fp);
			if (!colorchange ((char *) input, last))
				if (!cmdmatch (input))
					tprintf ("@%s@", input);
		} else
#ifndef DEMO
			if ((c != 0x0d) && (c != 0x1a))
#endif
				tputc (c);
	}
	(void) fclose (fp);
	if (invalid)
		last[0] = 0;
}



int
getcolor (char *str)
{
	if (strcasecmp(str, "black") == 0)
		return COLOR_BLACK;
	if (strcasecmp(str, "red") == 0)
		return COLOR_RED;
	if (strcasecmp(str, "yellow") == 0)
		return COLOR_YELLOW;
	if (strcasecmp(str, "green") == 0)
		return COLOR_GREEN;
	if (strcasecmp(str, "blue") == 0)
		return COLOR_BLUE;
	if (strcasecmp(str, "magenta") == 0)
		return COLOR_MAGENTA;
	if (strcasecmp(str, "cyan") == 0)
		return COLOR_CYAN;
	if (strcasecmp(str, "white") == 0)
		return COLOR_WHITE;
	return -1;
}



/* Set terminal colors */
int
docolor (int argc OPTIONAL, char *argv[], void *p OPTIONAL)
{
int temp;
int back, fore;

	temp = getcolor (argv[1]);
	if (temp == -1)
		tprintf (nope, argv[1]);
	else {
		back = temp;
		temp = getcolor (argv[2]);
		if (temp == -1)
			tprintf (nope, argv[2]);
		else {
			fore = temp;
#ifdef SM_CURSES
			if (argc > 3)	{
				if (argv[3][0] == 's')	{
					if (argv[3][1] == 'p')	{	/* splitline colors */
						SPLITback = back;
						SPLITfore = fore;
						(void) init_pairs();
						return 0;
					} else if (argv[3][1] == 't') {	/* statline colors */
						STATback = back;
						STATfore = fore;
						(void) init_pairs();
						return 0;
					}
				}
				/* must be main colors */
			}
#endif
#ifndef SM_CURSES
			SYSback = back;
			SYSfore = fore;
#endif
#ifdef SM_CURSES
			setscreens (back, fore, 0);
#else
			setscreens (0, 0, 0);
#endif				
			textrefresh ();
		}
	}
	return 0;
}



void
setscreens (int back OPTIONAL, int fore OPTIONAL, int clr OPTIONAL)
{
int resetstatline = 0;

	if (!back && !fore)	{	/* if both black, use system colors */
		back = SYSback;
		fore = SYSfore;
		resetstatline = 1;
	}

#ifdef SM_CURSES
	/* reinit colors if back or fore change */
	if (SYSback != back || SYSfore != fore) {
		SYSback = back;
		SYSfore = fore;
		(void) init_pairs();
	}
#endif
	textbackground (back);
	textcolor (fore);
	textattr ((back << 4) | fore);		/*lint !e701 */

	if (clr)	{
		clrscr ();
		if (resetstatline && STATLINE)	{
			displayStatLine (1, 1, 0);
			displayStatLine (-1, 1, 0);
		}
	}
}


#ifdef DEMO
void
main (argc, argv)
int argc;
char *argv[];
{
	last[0] = last[1] = 0;
	if (argc == 1)
		exit (0);
	colorfile (argv[1], last);
}
#endif

