import java.io.*;

/**
 * Title:        Convert pdftotext
 * Description:  Converts the output of the pdftotext application, to keep each remaining line < 80 characters.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Brian A. Lantz
 * @version 1.0
 */

public class convert {

private static final int maxlen = 79;
private static int linecounter = 0;
private static int eatlines = 0;
private static boolean lastwasfeed = false;
private static boolean firstpage = true;
private static boolean lastblank = false;
private static int lastLineLen = 0;
private static int lastIndentLevel = 0;
private static int sourceLine = 0;
private static PrintWriter pw = null;
private static boolean startedTOC = false;
private static boolean finishedTOC = false;
private static boolean pendingBlank = false;
private static final String TOC = "Table of Contents";
private static boolean lastOutputBlank = false;

  private static void log (String data) {
//    System.err.println(data);
  }

  private static void println (String data) {
    if (data.equals(TOC))
      pw.println("\n\n");
    if (isMajorSection (data))
      pw.println ();
    pw.println(data);
    lastOutputBlank = (data.trim().length() == 0);
    log ("--> " + data);
  }

  private static String read (BufferedReader br, boolean cont) {
    String data = null;
    pendingBlank = false;
    log ("[read: cont=" + cont + "]");
    while (true)  {
      try {
        data = br.readLine ();
        linecounter++;
      } catch (IOException e) {
        data = null;
      }
      sourceLine++;
      log ("<" + sourceLine + "<" + data);
      if (eatlines != 0)  {
        eatlines--;
        log ("eating line");
        continue;
      }
      if (data == null)
        break;

      if (data.indexOf(12) != -1) {
        // eat ^L, the blank line after, and the header line following
        firstpage = false;
        if (lastwasfeed)
          eatlines = 1;
        else
          eatlines = 2;
        lastwasfeed = true;
        lastblank = true;
        log ("found EOP, eating " + eatlines + " lines");
        continue;
      }

      if (cont) {
        if (data.trim().length() == 0)  {
          log ("eating blank on cont");
          continue;
        }
        if (isNewSection (data))
          data = "\n" + data;
      }

      if (firstpage == false) {
        // only eat redundant blank lines on pages other than the first
        if (lastblank && data.trim().length() == 0) {
          log ("eating duplicate blank on non firstpage");
          continue;
        }
      }

      lastwasfeed = false;
      lastblank = false;
      if (data.trim().length() < 5 && data.trim().length() > 0) {
        // eat a page number
        eatlines = 1;
        log ("eating a page number");
        continue;
      }

      int ind= 0;
      ind = data.indexOf("..........");
      if (ind != -1) {
        // modify the table of contents line
        data = "  " + data.substring(0, ind);
        log ("modifying a TOC line");
      }

      if (data.trim().length() == 0) {
        lastblank = true;
        if (startedTOC && !finishedTOC) {
          println ("");
          log ("finishedTOC");
          finishedTOC = true;
        }
        if (cont == false) {
          pendingBlank = true;
          continue;
        }
      }

      if (linecounter < 10 || !data.startsWith("                    "))
        break;
      // otherwise, eat the line
      log ("eating the line, cause unknown!");
    }
    if (data == null)
      lastLineLen = 0;
    else
      lastLineLen = data.length();
    lastIndentLevel = getIndent (data);
    log (">" + sourceLine + ">" + data);
    if (pendingBlank) {
      log ("outputing pending blank");
      println ("");
    }
    return data;
  }

  private static int getIndent (String data) {
    int retval = 0;
    if (data == null)
      return 0;
    for (retval = 0; retval < data.length(); retval++)  {
      if (!Character.isWhitespace(data.charAt(retval)))
        break;
    }
    return retval;
  }

  private static boolean isIndented (String data) {
    if (data.startsWith("          "))
      return true;
    return false;
  }

  private static boolean isMajorSection (String data) {
    if (data.length() < 4)
      return false;
    if (Character.isLetterOrDigit(data.charAt(0)) && data.charAt(1) == '.' && data.charAt(2) == ' ')
      return true;
    return false;
  }

  private static boolean isNewSection (String data) {
    String temp = data.trim();
    if (data.startsWith(TOC)) {
        return true;
    }

    if (temp.length() < 4)
      return false;
    if (data.charAt(0) == '*')
      return true;
    if (isMajorSection (temp)) {
        startedTOC = true;
        log ("startedTOC");
        return true;
    }
    return false;
  }

  public static void main(String[] args) {
    if (args.length != 2) {
      System.out.println("Usage: convert <infile> <outfile>");
      System.exit(-1);
    }

    BufferedReader br = null;
    try {
      br = new BufferedReader (new FileReader (args[0]));
      pw = new PrintWriter (new FileWriter (args[1]));
    } catch (IOException ioe) {
      ioe.printStackTrace ();
      System.exit(-1);
    }

    String data = null;
    boolean readone = true;
    boolean printit = true;
    // read initial line
    data = read (br, false);
    while (data != null) {    // while not EOF
      readone = true;
      printit = true;

      if (isNewSection (data) && finishedTOC) {
        // add another blank line before sections
        if (!lastOutputBlank)
          println ("");
      }

      if (data.length() > maxlen) {
        int last = data.lastIndexOf(' ',maxlen);
        int next = last + 1;
        if (last == -1) {
          // no space on line
          last = maxlen - 1;
          next = maxlen;
        }
        String temp = data.substring(0, last);
        data = data.substring(next);
        // output first part of line
        println(temp);

        StringBuffer sb = new StringBuffer ();
        if (!temp.startsWith ("*"))  {    // if initial line didn't begin with "*"
          // add same amount of leading white space
          for (int k = 0; k < temp.length(); k++) {
            if (temp.charAt(k) == ' ')
              sb.append (' ');
            else
              break;
          }
        } else
          data = "  " + data;
        sb.append (data);

        if (sb.length() > maxlen) {
          data = sb.toString();
        } else {
          // now read next line, to possibly combine
          int lastI = lastIndentLevel;
          temp = read (br, true /*(lastLineLen > maxlen) */);
          if (temp.trim().length() == 0 || isNewSection (temp) || (isIndented (temp) && lastI != lastIndentLevel) || temp.charAt(0) == '\n')  {
            println (sb.toString());
            if (temp != null && temp.length() > 0 && temp.charAt(0) == '\n')
              data = temp.substring(1);
            else
              data = temp;
          } else {
            sb.append (' ');
            sb.append (temp.trim());
            data = sb.toString();
          }
        }
        readone = false;
        printit = false;
      }

      if (printit)
        println(data);
      if (readone)
        data = read (br, false);
    }
    try {
      pw.flush();
      pw.close();
      br.close();
    } catch (IOException ioe3) { }
  }
}
