/*****************************************************************************/

/*
 *      native.c  -- directly attach a native BSD Sock interface.
 *
 * This module is NOT Linux specific.
 *
 * Developed for TNOS (brian@lantz.com)
 *
 */

#if !defined(_lint)
static char rcsid[] OPTIONAL = "$Id: native.c,v 1.1.1.1 2004/02/09 21:51:25 blantz Exp $";
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <net/if.h>
#include "system.h"

#ifdef __CYGWIN__
#include <cygwin/in.h>
#endif

#undef LOCK
#define	_SOCKET_H
#define _SOCKADDR_H
#include "global.h"

#if defined(NATIVE_IF)

#include "iface.h"

#include "commands.h"
#include "devparam.h"
#include "trace.h"
#include "hardware.h"
#include "enet.h"

typedef void * HANDLE;
#include "pcap.h"

#if 0
#ifdef __CYGWIN__
#define socket cygwin_socket
#endif
#endif

/* native interface control block */
struct nativeif {
    struct iface *iface;

    int fd;                 /* File descriptor for raw */
    int fd_udp;             /* File descriptor for UDP */
    int fd_tcp;             /* File descriptor for TCP */

    struct mbuf *sndq;      /* Transmit queue */

    short oldflags;         /* used to restore the interrupt flags */
    int proto;              /* protocol to listen for */
    int promisc;            /* set interface to promiscious mode */
    int class;		/* device class (for dump) */

    long rxpkts;            /* receive packets */
    long txpkts;            /* transmit packets */
    long rxchar;            /* Received characters */
    long txchar;            /* Transmitted characters */
    
    struct proc *rx_raw;
//    struct proc *rx_udp;
    //    struct proc *rx_tcp;
    const u_char *packet;
    char errbuf[PCAP_ERRBUF_SIZE];
    struct pcap_pkthdr hdr;
    pcap_t *pcap;
};

static struct nativeif NativeIF;

char *devname = "\\Device\\Packet_{046DB2CC-E863-4407-9D18-D46274238B1B}";
static int synced = 0;

static void nativeif_rx_raw(int dev, void *v1, void *v2);
//static void nativeif_rx_tcp(int dev, void *v1, void *v2);
//static void nativeif_rx_udp(int dev, void *v1, void *v2);


/*---------------------------------------------------------------------------*/

int nativeif_init (void)
{
    tprintf ("Using Network Device: %s\n", devname);
    NativeIF.pcap = pcap_open_live (devname, BUFSIZ, 1, 100, NativeIF.errbuf);
    if (NativeIF.pcap == NULL) {
        tprintf ("Error opening device: %s\n\t%s\n", devname, NativeIF.errbuf);
        return -1;
    }
    return 0;
}

static int nativeif_up(void)
{
//struct ifreq ifr;
//struct sockaddr sa;
char *cp;

//    if (NativeIF.fd >= 0 || NativeIF.fd_udp >= 0 || NativeIF.fd_tcp >= 0)        /* Already UP */
//        return 0;
    
    nativeif_init ();
    /*
    strcpy(ifr.ifr_name, ki->iface->name);
    strcpy(sa.sa_data, ki->iface->name);
    sa.sa_family = AF_INET;
    if (bind(ki->fd, &sa, sizeof(struct sockaddr)) < 0)
       goto Fail;
     */

    register_io (NativeIF.fd, &NativeIF.fd);
    register_io (NativeIF.fd_udp, &NativeIF.fd_udp);
    register_io (NativeIF.fd_tcp, &NativeIF.fd_tcp);

    cp = if_name (NativeIF.iface, " rx raw");
    NativeIF.rx_raw = newproc (cp, 768, nativeif_rx_raw, 0, NULL, NULL, 0);
//    cp = if_name (NativeIF.iface, " rx tcp");
//    NativeIF.rx_tcp = newproc (cp, 768, nativeif_rx_tcp, 0, NULL, NULL, 0);
//    cp = if_name (NativeIF.iface, " rx udp");
//    NativeIF.rx_udp = newproc (cp, 768, nativeif_rx_udp, 0, NULL, NULL, 0);
    tprintf ("Native IO interface should be active!\n");
    return 0;
}

/*---------------------------------------------------------------------------*/

static int nativeif_down(void)
{
    if (NativeIF.fd < 0)         /* Already DOWN */
        return 0;

    unregister_io (NativeIF.fd);
    unregister_io (NativeIF.fd_udp);
    unregister_io (NativeIF.fd_tcp);
    free_q(&NativeIF.sndq);
    killproc (NativeIF.rx_raw);
//    killproc (NativeIF.rx_tcp);
//    killproc (NativeIF.rx_udp);
    NativeIF.rx_raw = NULLPROC;
//    NativeIF.rx_udp = NULLPROC;
//    NativeIF.rx_tcp = NULLPROC;
//    close(NativeIF.fd);
//    close(NativeIF.fd_udp);
//    close(NativeIF.fd_tcp);
//    NativeIF.fd = -1;
//    NativeIF.fd_tcp = -1;
//    NativeIF.fd_udp = -1;
    tprintf ("Native IO interface should NOT be active!\n");
    return 0;
}

/*---------------------------------------------------------------------------*/

/* Asynchronous line I/O control */

static int32 nativeif_ioctl(struct iface *ifp OPTIONAL, int cmd, int set OPTIONAL, int32 val OPTIONAL)
{
    switch(cmd){
        case PARAM_DOWN:        return nativeif_down() ? 0 : 1;
	case PARAM_UP:		return nativeif_up() ? 0 : 1;
        default:	        return -1;
	}
}

/*---------------------------------------------------------------------------*/

/*
 * Send routine for packet driver
 */

static int
nativeif_send(
struct mbuf *bp,		/* Buffer to send */
struct iface *iface,		/* Pointer to interface control block */
uint32 gateway OPTIONAL,	/* Ignored  */
int prec OPTIONAL,
int del OPTIONAL,
int tput OPTIONAL,
int rel OPTIONAL
){
	if (iface == NULL)	{
		free_p (bp);
		return -1;
	}
	return (*iface->raw) (iface, bp);
}



static int nativeif_raw(struct iface *iface, struct mbuf *bp)
{
struct sockaddr to;
uint8 buf[4096]; /* should be enough */
uint8 *bufp = buf;
struct mbuf *bp2;
int cnt = 0;
int i;

    dump (iface, IF_TRACE_OUT, (unsigned) NativeIF.class, bp);

    iface->rawsndcnt++;
    iface->lastsent = secclock();
    if (iface->trace & IF_TRACE_RAW)
	raw_dump(iface, -1, bp);

    if (NativeIF.iface == NULL || NativeIF.fd < 0)
        free_p (bp);
    else {
	for (bp2 = bp; bp2; bp2 = bp2->next) {
            cnt += bp2->cnt;
            if (cnt > (int) sizeof(buf)) {
		(void) free_mbuf (bp);
		return -1;
            }
            memcpy (bufp, bp2->data, bp2->cnt);
            bufp += bp2->cnt;
	}
	strncpy(to.sa_data, NativeIF.iface->name, sizeof(to.sa_data));
			
#if 1
	i = sendto(NativeIF.fd, buf, (unsigned) cnt, 0, &to, sizeof(to));
#else
        i = cnt;
#endif
fprintf (stderr, "i=%d, errno=%d\n", i, errno);
tprintf ("i=%d, errno=%d\n", i, errno);

	if (i >= 0) {
            NativeIF.txpkts++;
            NativeIF.txchar += i;
            (void) free_mbuf(bp);
            return 0;
	}
	if (errno == EMSGSIZE) {
            (void) free_mbuf(bp);
            return -1;
	}
	if (errno == EWOULDBLOCK)
            return 0;
	(void) nativeif_down();
    }
    return 0;
}

/*---------------------------------------------------------------------------*/

static int nativeif_rx_handler (int fd)
{
struct sockaddr from;
int from_len = sizeof(from);
int i;
struct mbuf *bp;
    
    NativeIF.packet = pcap_next (NativeIF.pcap, &NativeIF.hdr);
    if (NativeIF.packet == NULL)
        return 0;
    
    if (synced == 0)
        if (NativeIF.packet[0] != 1)
            return 0;
    
    if ((bp = alloc_mbuf(NativeIF.hdr.len + 16)) == 0)
//  if ((bp = alloc_mbuf(NativeIF.iface->mtu+16)) == 0)
        return -1;
    i = NativeIF.hdr.len;
    memcpy (bp->data, NativeIF.packet, i);
//  i = recvfrom(fd, bp->data, bp->size, 0, &from, &from_len);
#if 0
    if (i <= 0) {
	if (i < 0 || errno != EWOULDBLOCK)
            (void) nativeif_down();
	(void) free_mbuf(bp);
	return -1;
    }
#endif
    synced = 1;
    bp->cnt = (int16) i;
    NativeIF.rxpkts++;
    NativeIF.rxchar += i;
    if (NativeIF.iface->trace & IF_TRACE_RAW)
	raw_dump(NativeIF.iface, IF_TRACE_IN, bp);

    (void) net_route(NativeIF.iface, NativeIF.class, bp);
    return 1;
}

static void nativeif_rx_raw (int i OPTIONAL, void *v1 OPTIONAL, void *v2 OPTIONAL)
{
    int retval = 1;
    server_disconnect_io ();

    for ( ; ; )	{
        //      	kwait (&NativeIF.fd);
        if (retval == 0)
            kpause (__FILE__, __LINE__, 100);
        retval = nativeif_rx_handler (NativeIF.fd);
        if (retval == -1)
            break;
    }
    NativeIF.rx_raw = NULLPROC;
}

#if 0
static void nativeif_rx_udp (int dev OPTIONAL, void *v1 OPTIONAL, void *v2 OPTIONAL)
{
    server_disconnect_io ();

    for ( ; ; )	{
	kwait (&NativeIF.fd_udp);
        if (nativeif_rx_handler (NativeIF.fd_udp))
            break;
    }
    NativeIF.rx_udp = NULLPROC;
}

static void nativeif_rx_tcp (int dev OPTIONAL, void *v1 OPTIONAL, void *v2 OPTIONAL)
{
    server_disconnect_io ();

    for ( ; ; )	{
	kwait (&NativeIF.fd_tcp);
        if (nativeif_rx_handler (NativeIF.fd_tcp))
            break;
    }
    NativeIF.rx_tcp = NULLPROC;
}
#endif

/*---------------------------------------------------------------------------*/

static void nativeif_status(struct iface *iface)
{
    if (NativeIF.iface != iface)
        return;

    tprintf("           Received:    Packets %8ld Chars %8ld\n"
	       "           Transmitted: Packets %8ld Chars %8ld\n",
	       NativeIF.rxpkts, NativeIF.rxchar, NativeIF.txpkts, NativeIF.txchar);
}

/*---------------------------------------------------------------------------*/

static int nativeif_detach(struct iface *ifp)
{
    if (ifp == NULL)
	return -1;
    if(NativeIF.iface == NULL)
	return -1;      /* Not allocated */
    (void) nativeif_down();
    NativeIF.iface = NULL;
    return 0;
}

/*---------------------------------------------------------------------------*/

int nativeif_attach(int argc OPTIONAL, char *argv[], void *p OPTIONAL)
{
struct iface *ifp;
//int dev;
//struct ifreq ifr;
//struct sockaddr hw;
//struct sockaddr_in in;
//int fd;

    if (if_lookup(argv[1]) != NULL) {
		tprintf("Interface %s already exists\n",argv[1]);
		return -1;
	}
	/*
	 * get parameters of the interface
	 */
    
	/* Create interface structure and fill in details */
	ifp = (struct iface *)callocw(1,sizeof(struct iface));
	ifp->addr = Ip_addr;
	ifp->name = strdup(argv[1]);
	ifp->mtu = MAXINT16;
	ifp->dev = 0;
	ifp->stop = nativeif_detach;
        /* set encapsulation */
    (void) setencap (ifp, /*"None" */ "Ethernet");
    NativeIF.class = 1;
	ifp->ioctl = nativeif_ioctl;
	ifp->raw = nativeif_raw;
        ifp->show = nativeif_status;
        ifp->send = nativeif_send;
	ifp->xdev = 0;
	
	NativeIF.proto = 0;
        NativeIF.promisc = 1;
    
//	NativeIF.fd = -1;
//	NativeIF.fd_udp = -1;
//	NativeIF.fd_tcp = -1;
	NativeIF.iface = ifp;

	/* Link in the interface */
	ifp->next = Ifaces;
	Ifaces = ifp;

	return nativeif_up();
}

/*---------------------------------------------------------------------------*/


#endif /* NATIVE_IF */
