#include "global.h"
#ifdef SHELL
#include <signal.h>
#include "proc.h"
#include "hardware.h"
#include "session.h"
#ifdef	T3_ISHELL
/* 19Apr2006, Maiko (VE4KLM), Make the 'Hit enter to continue' an option */
extern int keywait (const char *prompt,int flush);
#endif
extern int sockblock (int s,int value);
extern int recvchar (int s);
extern int recvline (int s,unsigned char *buf,unsigned len);
extern void tflush (void);
#define setflush j_setflush
extern int setflush (int s,int c);
extern int socketalive (int s);
extern int dup (int oldfd);
extern void displayStatLine (int offset, int phase, int onlymarquee);
#include <sys/wait.h>
#include <sys/time.h>

#if !defined(_lint)
static char rcsid[] OPTIONAL = "$Id: pipe.c,v 1.2 2006/04/19 14:18:26 mlangelaar Exp $";
#endif

#ifdef SOCK_STREAM
#undef SOCK_STREAM
#endif
#include <sys/socket.h>

#ifdef socketpair
#undef socketpair
#endif

static int s_pipe(int fd[2]);
static void doshell_interactive (int argc, char *argv[], void *p);
int doshell (int argc,char *argv[],void *p);
static void shell_io(int fd, pid_t pid);
extern int dup2(int oldfd, int newfd);

extern unsigned char SCREENwidth, SCREENlength;

#define STDIN_FILENO 0
#define STDOUT_FILENO 1
#define STDERR_FILENO 2
#define MAXLINE 512

#define SOCK_NOTXBLOCK  1
#define SOCK_NORXBLOCK  2
#ifndef _lint
#define SOCK_BINARY 0
#endif

static const char strCR[] = "shell: %s\n";


static void
shell_io (int fd, pid_t pid)
{
int n;
char line[MAXLINE];
int retval;
int oldblock = 0;
int oldblock2 = 0;
int oldflush = 0;
int oldflush2 = 0;
fd_set fdr;
struct timeval tv;
struct timeval *tvp;
int fd2;

	tflush ();
	oldblock = sockblock(Curproc->input,SOCK_NORXBLOCK);
	oldblock2 = sockblock(Curproc->output,SOCK_NOTXBLOCK);
	oldflush = setflush (Curproc->input, -1);
	oldflush2 = setflush (Curproc->output, -1);

	fd2 = dup (fd);
	tvp = &tv;

	kwait (NULL);
	do	{
		if (!socketalive (Curproc->input))
			break;

		kwait (NULL);
		retval = recvchar(Curproc->input);
		if (retval == EOF && errno != EWOULDBLOCK)
			break;
		else	if (retval != EOF)	{
				line[0] = (char) retval;
				kwait (NULL);
				if (write (fd, line, 1) != 1)	{
					tprintf (strCR, "write error");
					break;
				}
		}

		kwait (NULL);
		tv.tv_sec = tv.tv_usec = 0;
		FD_ZERO(&fdr);
		FD_SET(fd2, &fdr);
		if (select(fd2 + 1, &fdr, 0, 0, tvp))	{
			n = read (fd2, line, MAXLINE);
			if (n > 0) {
				line[n] = 0;
				tputs(line);
			}
		}
		kwait (NULL);
		if (waitpid (pid, NULL, WNOHANG))
			break;

		kwait (NULL);
	} while (1);

	while ((n = read (fd2, line, MAXLINE)) > 0)	{
		line[n] = 0;
		tputs (line);
		kwait (NULL);
	}
	close (fd2);
	tputs("\n");

	(void) sockblock(Curproc->input,oldblock);
	(void) sockblock(Curproc->output,oldblock2);
	(void) setflush(Curproc->input,oldflush);
	(void) setflush(Curproc->output,oldflush2);
}


static void
shell_cmd (int argc OPTIONAL, char *argv[], void *p OPTIONAL)
{
int fd[2];
struct session *sp = NULLSESSION;
int usesession = 0;
pid_t pid;
int manage_session = (int) p;

	/* Use a session if this comes from console */
	if(manage_session && Curproc->input == Command->input) {
		usesession = 1;
		if((sp = newsession(argv[2],SHSESSION,0)) == NULLSESSION)	{
			tputs (TooManySessions);
			return;
		}
		if (sp->screen->statline == 0)	{
			sp->screen->statline = 1;
			displayStatLine (0, 0, 0);
		}
		chname (Curproc, "Shell command");
	}

	(void) signal (SIGPIPE, SIG_IGN);
	if (s_pipe (fd) < 0)	{
		tprintf (strCR, "shell error");
		return;
	}

	switch (pid = fork())	{
		case -1:	tprintf (strCR,"fork error");
				return;

		case 0:		/* child */
			close (fd[0]);
			if (fd[1] != STDIN_FILENO)	{
				if (dup2 (fd[1], STDIN_FILENO) != STDIN_FILENO)	{
					tprintf(strCR, "dup2 error to stdin");
					return;
				}
			}
			if (fd[1] != STDOUT_FILENO)	{
				if (dup2 (fd[1], STDOUT_FILENO) != STDOUT_FILENO)	{
					tprintf(strCR, "dup2 error to stdout");
					return;
				}
			}
			if (fd[1] != STDERR_FILENO)	{
				if (dup2 (fd[1], STDERR_FILENO) != STDERR_FILENO)	{
					tprintf(strCR, "dup2 error to stderr");
					return;
				}
			}
			if (execvp (argv[0], (char * const *) argv) < 0)	{
				tprintf(strCR, "execvp error");
				return;
			}
			break;

		default:	/* parent */
			close (fd[1]);
			shell_io(fd[0], pid);
			close (fd[0]);
			(void) waitpid (pid, NULL, 0);
			(void) signal (SIGPIPE, SIG_DFL);

			if (usesession)
			{
#ifdef	T3_ISHELL
				(void) keywait(NULLCHAR,1);
#endif
				freesession(sp);
			}
			return;
	}
}


static int
s_pipe (int fd[2])
{
	return (socketpair (AF_UNIX, SOCK_STREAM, 0, fd));
}


static void
doshell_interactive (int argc OPTIONAL, char *argv[] OPTIONAL, void *p OPTIONAL)
{
int usesession = 0;
struct session *sp = NULLSESSION;
char line[256];
char *newargv[4];
int newargc = 3;
int n;

	/* Use a session if this comes from console */
	if(Curproc->input == Command->input) {
		usesession = 1;
		if((sp = newsession("shell",SHSESSION,0)) == NULLSESSION)	{
			tputs (TooManySessions);
			return;
		}
		if (sp->screen->statline == 0)	{
			sp->screen->statline = 1;
			displayStatLine (0, 0, 0);
		}
		chname (Curproc, "Shell interpretter");
	}

	newargv[0] = (char *) "/bin/sh";
	newargv[1] = (char *) "-c";
	newargv[2] = line;
	newargv[3] = NULLCHAR;
	
	while (1)	{
		tputs ("\nTNOS> ");
		tflush ();
		n = recvline (Curproc->input, line, 256);
		if (n > 0)	{
			rip (line);
			if (!strcmp (line, "exit") || !strncmp (line, "ex", 2))
				break;
			shell_cmd (newargc, newargv, 0);
		} else
			break;
	}


	tputs("\n");
	if (usesession)
	{
#ifdef	T3_ISHELL
		(void) keywait(NULLCHAR,1);
#endif
		freesession(sp);
	}
}


int
doshell (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
char **newargv;
int i;
int newargc = 3;
char line[256];

	if (!--argc)	{
		if(Curproc->input == Command->input)	/* spawn off subprocess and return */
		        (void) newproc("shell",512,(void (*)(int,void *,void *))doshell_interactive,0,NULLCHARP,0,0);
		else
			doshell_interactive (0, NULLCHARP, 0);
		return 0;
	}

	argv++;				/*lint !e608 */

	/* Make private copy of argv and args, modifying */
	newargv = (char **)callocw((size_t)(newargc + 1),sizeof(char *));
	if ((newargv[0] = getenv ("SHELL")) == NULLCHAR)
		newargv[0] = strdup ("/bin/sh");

	newargv[1] = strdup ("-c");
	line[0] = 0;

	for(i = 0; i < argc; i++)	{
		if (*line)
			strcat (line, " ");
		strcat (line, argv[i]);
	}
	newargv[2] = strdup (line);
	newargv[3] = NULL;

	if(Curproc->input == Command->input)	/* spawn off subprocess and return */
	        (void) newproc("pipe",512,(void (*)(int,void *,void *))shell_cmd,newargc,(void *)newargv,(void *)1,1);
	else	{
		shell_cmd (newargc, newargv, (void *)1);
		for (i = 0; i < newargc; i++)
			free (newargv[i]);
	}
	return 0;
}

#endif	/* SHELL */

