/*
 * Support for the TUN device - now we can use TUN as an alternative
 * to SLIP, when interfacing NOS to linux or unix. This my attempt at
 * writing a TUN driver for NOS.
 *
 * Designed and Coded by Maiko Langelaar, VE4KLM. The prototype was done
 * on JNOS, got it working Nov 22, 2004. Ported to TNOS on Dec 7, 2004
 * by Maiko as well.
 *
 * For Amateur Radio use only (please) ! Also, if you plan on using this
 * code for any other platform then the JNOS 2.0 that I maintain, please
 * do NOT remove this comment section, leave it as is. Thank you.
 *
 * $Author: mlangelaar $
 *
 * $Date: 2005/01/18 17:01:30 $
 *
 */

#if !defined(_lint)
static char rcsid[] OPTIONAL = "$Id: tun.c,v 1.2 2005/01/18 17:01:30 mlangelaar Exp $";
#endif

#include "global.h"
#include "mbuf.h"
#include "iface.h"
#include "ip.h"
#include "slhc.h"
#include "asy.h"
#include "slip.h"
#include "trace.h"
#include "pktdrvr.h"

#include <stdio.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
 
#include <linux/ioctl.h>
#include <linux/if.h>
#include <linux/if_tun.h> 

/*
 * Check the kernel version, since TUN device naming convention differs
 * between version 2.2 and version 2.4 kernels.
 */
#include <linux/version.h>

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) (((a) << 16) + ((b) << 8) + (c))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
#define	TUNDEVNAME	"/dev/net/tun"
#define	LINUX_2_4
#else
#warning "using /dev/tun0 device for older kernel"
#define	TUNDEVNAME	"/dev/tun0"
#define	LINUX_OLD
#endif

/* 07Dec2004, Maiko, Stuff that came up when compiling for TNOS 3 */
#define	log tnoslog
#define	pwait kwait
#include "unix.h"
#include <fcntl.h>
#include <errno.h>
int tun_raw (struct iface *iface, struct mbuf *bp);
void tun_rx (int xdev, void *p1, void *p2);
int tun_attach (int argc, char *argv[], void *p);
/* end of Stuff */

static struct slip Tun;
  
static int tunfd = -1;

int tun_send (bp, iface, gateway, prec, del, tput, rel)
	struct mbuf *bp;
	struct iface *iface;
	int32 gateway;
	int prec;
	int del;
	int tput;
	int rel;
{
	if (iface == NULLIF)
	{
		free_p (bp);
		return -1;
	}

    return (*iface->raw)(iface, bp);
}

int tun_raw (struct iface *iface, struct mbuf *bp)
{
	static char *iobuf = NULL;
	static int iomtu = -1;
	struct mbuf *tbp;
	int len;

    dump (iface, IF_TRACE_OUT, CL_NONE, bp);

	/*
	 * This code was written based on my stripping down of the asy_input()
	 * function used by the SLIP interface (asy.c, unixasy.c, slip.c).
	 *
	 * I don't see the point of allocating memory before a read, just
	 * to free it again right after the read, and so on. All that does
	 * is fragment memory. Only allocate the io buffer once at startup,
	 * and only re-allocate if the MTU of the interface is changed to a
	 * larger value than what it started at (why that would be the case
	 * is beyond me, but we should check for that to be safe).
 	 *
	 * 26Nov2004, Maiko Langelaar, VE4KLM
	 */

	if (iomtu == -1 || iface->mtu > iomtu)
	{
		if (iomtu != -1)
			free (iobuf);

		iomtu = iface->mtu;
		iobuf = mallocw (iomtu);
	}

    iface->lastsent = secclock();

    iface->rawsndcnt++;

	if (tunfd != -1)
	{
		len = len_p (bp);

		/* this should never happen, but just to be safe !!! */
		if (len > iomtu)
			log (-1, "tun_raw - packet larger then mtu");

		else if (dup_p (&tbp, bp, 0, len) != len)
			log (-1, "tun_raw - dup_p failed");

		else
		{
			(void)pullup (&tbp, iobuf, len);

			write (tunfd, iobuf, len);

			free_p (tbp);
		}

		free_p (bp);
	}

    return 0;
}

void tun_rx (int xdev, void *p1, void *p2)
{
	/* NOTE : should really malloc the rcv buffer based on mtu of iface */
	static char buf[1600];
	struct mbuf *bp;
	fd_set fds;
	int len;

	/* we pass the interface parameters into here */
	struct iface *ifp = (struct iface*)p1;

	log (-1, "tun_rx - listening for packets");

	while (1)
	{
		/* wait for an event, let others run if there is no event */
		if (pwait (&tunfd) != 0)
			return;

		if ((len = read (tunfd, buf, sizeof(buf))) == 0)
			continue;

		if (len == -1 && errno == EWOULDBLOCK)
			continue;

		if (len == -1)
		{
			/* timeout, retry */
			if (errno == EINTR)
				continue;

			tprintf ("tun_rx : read errno %d, shutting down\n", errno);

			return;
		}

#ifdef	DONT_COMPILE

		log (-1, "tun_rx - %d bytes", len);

		{
			static char logbuf[400], *lptr;

			int cnt;

			for (lptr = logbuf, cnt = 0; cnt < len; cnt++)
				lptr += sprintf (lptr, " %02X", (int)buf[cnt]);

			log (-1, "tun_rx data -%s", logbuf);
		}
#endif
		bp = qdata (buf, len);

    	dump (ifp, IF_TRACE_IN, CL_NONE, bp);

        if (net_route (ifp, CL_NONE, bp) != 0)
		{
			log (-1, "tun_rx - net_route failed !!!");

			free_p (bp);
		}
	}
}

static void tun_status (struct iface *iface)
{
    struct slip *sp;
  
    sp = &Tun;

    if (sp->iface != iface)
        /* Must not be a SLIP device */
        return;
  
    tprintf("  IN:\t%lu pkts\n", iface->rawrecvcnt);
    tprintf("  OUT:\t%lu pkts\n", iface->rawsndcnt);
}

static int tun_detach (struct iface *ifp)
{
    // tun_stop (ifp);
  
	Tun.iface = NULLIF;

    return 0;
}

int tun_attach (int argc, char *argv[], void *p)
{
	struct iface *ifp;
  
	if (if_lookup (argv[1]) != NULLIF)
	{
		tprintf (Existingiface, argv[4]);
		return -1;
	}

	/* Create interface structure and fill in details */

	ifp = (struct iface*)callocw (1, sizeof(struct iface));

	ifp->addr = Ip_addr;
	ifp->name = strdup (argv[1]);
	ifp->mtu = atoi (argv[2]);
	ifp->dev = atoi (argv[3]);
	ifp->stop = tun_detach;

	setencap (ifp, "TUN");

	ifp->ioctl = NULL;
	ifp->raw = tun_raw;
	ifp->show = tun_status;
	ifp->flags = 0;
	ifp->xdev = 0;

	/* Link in the interface - important part !!! */
	ifp->next = Ifaces;
	Ifaces = ifp;

	/* Now physically open TUN device to get things going */

	{
		struct sigaction sa;

  		struct ifreq ifr;
 
	    if ((tunfd = open(TUNDEVNAME, O_RDWR|O_NONBLOCK|O_NOCTTY, 0644)) == -1)
		{
			log (-1, "unable to open [%s]", TUNDEVNAME);
			return -1;
		}
 
#ifdef	LINUX_2_4

  		memset (&ifr, 0, sizeof(ifr));

  		ifr.ifr_flags = IFF_TUN | IFF_NO_PI;

  		if (ioctl (tunfd, TUNSETIFF, &ifr) < 0)
		{
			log (-1, "unable to setup tunnel");
			close (tunfd);
			return -1;
		}
 
  		log (-1, "using new [%s] device", ifr.ifr_name);
#else

		if (ioctl (tunfd, TUNSETNOCSUM, 1) < 0)
		{
			log (-1, "unable to setup tunnel");
			close (tunfd);
			return -1;
		}

  		log (-1, "using [%s] device", TUNDEVNAME);
#endif

	}

	ifp->rxproc = newproc ("tun_rx", 1024, tun_rx, 0, (void*)ifp, NULL, 0);

	/* pwait() will never return if we don't register the fd !!! */
	register_io(tunfd, &tunfd);

    return 0;
}

